import kamzik3
from kamzik3 import units
from kamzik3.constants import *
from kamzik3.devices.attribute import Attribute
from kamzik3.gui.deviceDebugWidget import DeviceDebugWidget
from kamzik3.gui.general.macroWidget import MacroWidget
from kamzik3.gui.templates.macroStepTemplate import Ui_Form
from kamzik3.macro.step import StepDeviceMethod, StepDeviceAttributeNumerical, StepDeviceAttribute
from kamzik3.snippets.snippetsWidgets import show_centralized_top, show_error_message


class MacroStepWidget(Ui_Form, MacroWidget):
    device = None
    attribute = None
    debug_widget = None

    def __init__(self, title=u"Step settings", config=None, parent=None):
        MacroWidget.__init__(self, parent)
        self.config = config
        if self.config is None:
            self.config = {}
        self.setupUi(self)
        if self.config.get("step") is not None:
            self.combobox_step.hide()
            self.label_step.hide()

        self.button_settings.setEnabled(False)
        self.method_attributes = []
        self.set_title(title)
        self.spacer = self.gridLayout.itemAtPosition(7, 0)
        self.label_attribute.hide()
        self.combobox_attribute.hide()
        self.label_method.hide()
        self.combobox_method.hide()
        self._init_device_list()
        self._init_scanner_list()
        self._init_template_settings()

    def _init_device_list(self):
        self.combobox_device.blockSignals(True)
        if self.config.get("attribute") is not None:
            attribute = self.config.get("attribute").split(u" # ")
            devices_list = list(kamzik3.session.get_devices(attribute_filter=attribute).keys())
        elif self.config.get("method") is not None:
            devices_list = list(kamzik3.session.get_devices(method_filter=self.config.get("method")).keys())
        else:
            devices_list = list(kamzik3.session.get_devices().keys())

        self.combobox_device.addItems(["None"] + devices_list)
        self.combobox_device.blockSignals(False)

        if self.config.get("device_id") is not None:
            self.combobox_device.setCurrentText(self.config.get("device_id"))

    def _init_step_list(self, device):
        current_text = self.combobox_step.currentText()
        self.combobox_step.clear()
        self.combobox_step.blockSignals(True)
        if device is None:
            self.combobox_step.addItem("None")
        else:
            attribute_list = []
            self.get_devices_macro_steps(device.macro_steps, attribute_list)
            self.combobox_step.addItems(attribute_list)
        index = self.combobox_step.findText(current_text)
        if index == -1:
            self.combobox_step.setCurrentIndex(0)
        else:
            self.combobox_step.setCurrentIndex(index)
        self.combobox_step.blockSignals(False)
        self.step_selected(self.combobox_step.currentText())

    def _init_attribute_list(self, device):
        current_text = self.combobox_attribute.currentText()
        self.combobox_attribute.clear()
        self.combobox_attribute.blockSignals(True)
        if device is None:
            self.combobox_attribute.addItem("None")
        else:
            attribute_list = []
            self.get_settable_attribute_list(device.attributes, attribute_list)
            if len(attribute_list) == 0:
                attribute_list.append("None")
            self.combobox_attribute.addItems(attribute_list)
        index = self.combobox_attribute.findText(current_text)
        if index == -1:
            self.combobox_attribute.setCurrentIndex(0)
        else:
            self.combobox_attribute.setCurrentIndex(index)
        self.combobox_attribute.blockSignals(False)
        self.attribute_selected(self.combobox_attribute.currentText())

    def _init_method_list(self, device):
        current_text = self.combobox_method.currentText()
        self.combobox_method.clear()
        self.combobox_method.blockSignals(True)
        if device is None:
            self.combobox_method.addItem("None")
        else:
            methods_list = []
            self.get_device_methods_list(device.exposed_methods, methods_list)
            if len(methods_list) == 0:
                methods_list.append("None")
            self.combobox_method.addItems(methods_list)
        index = self.combobox_method.findText(current_text)
        if index == -1:
            self.combobox_method.setCurrentIndex(0)
        else:
            self.combobox_method.setCurrentIndex(index)
        self.combobox_method.blockSignals(False)
        self.method_selected(self.combobox_method.currentText())

    def _init_template_settings(self):
        if not self.config.get(CFG_TEMPLATE, False):
            self.label_settings.hide()
            self.widget_settings.hide()
            self.label_scanner.hide()
            self.combobox_scanner.hide()

    def _init_scanner_list(self):
        self.combobox_scanner.blockSignals(True)
        devices_list = list(kamzik3.session.get_devices(method_filter="get_scanner_macro").keys())

        self.combobox_scanner.addItems(["None"] + devices_list)
        self.combobox_scanner.blockSignals(False)

        if self.config.get("device_id") is not None:
            self.combobox_scanner.setCurrentText(self.config.get("device_id"))

    def set_title(self, title):
        self.title = title
        self.label_axis_settings.setText(title)

    def device_selected(self, device_id):
        if device_id == "None":
            self.device = None
            self._init_step_list(None)
            self.button_settings.setEnabled(False)
        else:
            self.button_settings.setEnabled(True)
            self.device = kamzik3.session.get_device(device_id)
            if self.config.get("attribute") is not None:
                self.attribute_selected(self.config.get("attribute"))
            elif self.config.get("method") is not None:
                self.method_selected(self.config.get("method"))
            else:
                self._init_step_list(self.device)

    def step_selected(self, step_id):
        if step_id == MACRO_SET_ATTRIBUTE_STEP:
            self._init_attribute_list(self.device)
            self.combobox_method.hide()
            self.label_method.hide()
            self.combobox_attribute.show()
            self.label_attribute.show()
        elif step_id == MACRO_EXECUTE_METHOD_STEP:
            self._init_method_list(self.device)
            self.combobox_attribute.hide()
            self.label_attribute.hide()
            self.combobox_method.show()
            self.label_method.show()
        else:
            self.combobox_attribute.hide()
            self.label_attribute.hide()
            self.combobox_method.hide()
            self.label_method.hide()

    def attribute_selected(self, attribute):
        if self.config.get(CFG_TEMPLATE, False):
            return

        for widget in self.macro_widgets:
            widget.close()
            widget.setParent(None)
        self.macro_widgets = []
        self.macro_inputs = {}
        if attribute == "":
            return

        attribute = attribute.split(u" # ")
        self.attribute = self.device.get_attribute(attribute)
        row = 5
        self.macro_widgets.append(self._show_attribute_widget(row))
        row += 1
        if self.attribute.numerical:
            self.macro_widgets.append(self._show_negative_tolerance_widget(row))
            row += 1
            self.macro_widgets.append(self._show_positive_tolerance_widget(row))
            row += 1
            # self.macro_widgets.append(self._show_random_deviation_widget(row))
            # row += 1
        self.macro_widgets.append(self._show_wait_after_widget(row))
        row += 1
        self.macro_widgets.append(self._show_repeat_widget(row))
        row += 1
        self.macro_widgets.append(self._show_retry_widget(row))
        row += 1
        if self.config.get("scanner", "None") != "None":
            scanner = kamzik3.session.get_device(self.config["scanner"])
            scanner_attributes = scanner.get_scanner_attributes()
            for attribute in scanner_attributes:
                self.macro_widgets.append(
                    self._show_method_attribute_widget(attribute, scanner.get_attribute(attribute).attribute_copy(),
                                                       row))
                row += 1
        row += 1
        self.layout().removeItem(self.spacer)
        self.layout().addItem(self.spacer, row, 0)

    def method_selected(self, method):
        if self.config.get(CFG_TEMPLATE, False):
            return

        for widget in self.macro_widgets:
            widget.close()
            widget.setParent(None)
        self.macro_widgets = []
        self.macro_inputs = {}
        self.method_attributes = []
        if method == "":
            return

        row = 5
        for exposed_method in self.device.exposed_methods:
            if exposed_method[0] == method:
                if exposed_method[1] is not None:
                    for attribute_title, attribute_type in exposed_method[1].items():
                        device = kamzik3.session.get_device(self.combobox_device.currentText())
                        device_attribute = device.get_attribute(attribute_type)
                        self.method_attributes.append(attribute_title)
                        self.macro_widgets.append(
                            self._show_method_attribute_widget(attribute_title, device_attribute, row))
                        row += 1
                break
        self.macro_widgets.append(self._show_wait_after_widget(row))
        row += 1
        self.macro_widgets.append(self._show_repeat_widget(row))
        row += 1
        self.macro_widgets.append(self._show_retry_widget(row))
        row += 1
        self.macro_widgets.append(self._show_busy_timeout(row))
        row += 1
        if self.config.get("scanner", "None") != "None":
            scanner = kamzik3.session.get_device(self.config["scanner"])
            scanner_attributes = scanner.get_scanner_attributes()
            for attribute in scanner_attributes:
                self.macro_widgets.append(
                    self._show_method_attribute_widget(attribute, scanner.get_attribute(attribute).attribute_copy(),
                                                       row))
                row += 1
        self.layout().removeItem(self.spacer)
        self.layout().addItem(self.spacer, row, 0)

    def get_settable_attribute_list(self, attributes, output=[]):
        for k, v in attributes.items():
            if Attribute.is_attribute(v):
                if not v[READONLY]:
                    output.append(k)
            else:
                leaf = []
                self.get_settable_attribute_list(v, leaf)
                output += map(lambda nk: "{} # {}".format(k, nk), leaf)

    @staticmethod
    def get_devices_macro_steps(macro_steps, output=[]):
        for step_title in macro_steps.keys():
            output.append(step_title)

    @staticmethod
    def get_device_methods_list(device_methods, output=[]):
        for method in device_methods:
            output.append(method[0])

    def get_template(self):
        template = {
            "type": "action",
            "device_id": self.combobox_device.currentText(),
            "step": self.combobox_step.currentText(),
            "trigger_log": self.checkbox_log.isChecked(),
            "scanner": self.combobox_scanner.currentText(),
            "defaults": self.checkbox_defaults.isChecked(),
        }
        if template["step"] == "Set attribute":
            template["attribute"] = self.combobox_attribute.currentText()
        elif template["step"] == "Execute method":
            template["method"] = self.combobox_method.currentText()
        return template

    def save(self):
        if self.combobox_device.currentText() == "None":
            raise KeyError("Device was not set for {}".format(self.title))
        try:
            values = self.get_input_values()
            step = None
            if self.config.get("attribute") is not None:
                if self.attribute.numerical:
                    step = StepDeviceAttributeNumerical(
                        common_id="Step",
                        device_id=self.combobox_device.currentText(),
                        attribute=self.config.get("attribute").split(u" # "),
                        setpoint=values["value"],
                        negative_tolerance=values["positive_tolerance"] if self.attribute.numerical else None,
                        positive_tolerance=values["negative_tolerance"] if self.attribute.numerical else None,
                        repeat_count=values["repeat"],
                        wait_after=values["wait_after"],
                        retry_count=int(values["retry_count"].m),
                        trigger_log=self.config.get("trigger_log", True)
                    )
                else:
                    step = StepDeviceAttribute(
                        common_id="Step",
                        device_id=self.combobox_device.currentText(),
                        attribute=self.config.get("attribute").split(u" # "),
                        setpoint=values["value"],
                        repeat_count=values["repeat"],
                        wait_after=values["wait_after"],
                        retry_count=int(values["retry_count"].m),
                        trigger_log=self.config.get("trigger_log", True)
                    )
            elif self.config.get("method") is not None:
                method_parameters = {}
                for method_attribute in self.method_attributes:

                    if isinstance(values[method_attribute], units.Quantity):
                        method_parameters[method_attribute] = "{:~}".format(values[method_attribute])
                    else:
                        method_parameters[method_attribute] = values[method_attribute]
                step = StepDeviceMethod(
                    common_id="Step",
                    device_id=self.combobox_device.currentText(),
                    method=self.config.get("method"),
                    method_parameters=method_parameters,
                    repeat_count=values["repeat"],
                    wait_after=values["wait_after"],
                    retry_count=int(values["retry_count"].m),
                    busy_timeout=values["busy_timeout"],
                    trigger_log=self.config.get("trigger_log", True)
                )
            if self.config.get("scanner", "None") != "None":
                scanner = kamzik3.session.get_device(self.config["scanner"])
                scanner_attributes = {}
                for attribute in scanner.get_scanner_attributes():
                    scanner_attributes[attribute] = self.macro_inputs[attribute].get_attribute_value()
                step = scanner.get_scanner_macro(step, scanner_attributes)
            return step
        except KeyError:
            raise KeyError("One of attributes for {} were not set".format(self.title))

    def show_debug_widget(self):
        if self.device is None:
            show_error_message("No device selected", parent=self)
            return
        if self.debug_widget is None:
            self.debug_widget = DeviceDebugWidget(self.device)
        elif self.debug_widget.device != self.device:
            self.debug_widget.close()
            self.debug_widget.setParent(None)
            self.debug_widget.deleteLater()
            self.debug_widget = None
            self.debug_widget = DeviceDebugWidget(self.device)

        show_centralized_top(self.debug_widget)



import numpy as np

import kamzik3
from kamzik3 import MacroException
from kamzik3 import session
from kamzik3.constants import *
from kamzik3.devices.general.deviceScanner import DeviceScanner
from kamzik3.macro.macro import Macro
from kamzik3.macro.step import StepDeviceAttributeNumerical
from kamzik3.snippets.snippetsDecorators import expose_method


class DummyScanner(DeviceScanner):

    def _init_attributes(self):
        DeviceScanner._init_attributes(self)
        self.create_attribute(ATTR_DWELL_TIME, default_value=100, unit="ms", min_value=0, default_type=np.int64)

    def _init_new_scan(self, scanner_input, scanner_attributes):
        self.logger.info(u"Initiating new scan number {}".format(kamzik3.session.get_value(ATTR_SCAN_COUNT)))

    @expose_method()
    def get_scanner_attributes(self):
        return [ATTR_DWELL_TIME]

    @expose_method()
    def get_scanner_macro(self, scanner_input, scanner_attributes):
        macro = Macro(device_id="{}_{}".format(self.device_id, scanner_input.common_id))
        distance = abs(scanner_input.start_point - scanner_input.end_point)
        dwell_time = scanner_attributes[ATTR_DWELL_TIME]
        device_id = scanner_input.step_attributes['device_id']
        # current_velocity = session.get_device(device_id).get_attribute(ATTR_MAXIMUM_VELOCITY).value()
        maximum_velocity = session.get_device(device_id).get_attribute(ATTR_MAXIMUM_VELOCITY).maximum()
        mv = session.get_device(device_id).get_attribute(ATTR_MAXIMUM_VELOCITY)
        scan_line_time = dwell_time * scanner_input.steps_count
        final_velocity = distance / scan_line_time
        final_velocity = final_velocity.to(maximum_velocity.u)
        if final_velocity > maximum_velocity:
            raise MacroException(
                "Calculated velocity {} is higher than maximum velocity {}.".format(final_velocity, maximum_velocity))

        scanner_input.set_steps_count(1)
        if not scanner_input.bidirectional:
            velocity_step_0 = StepDeviceAttributeNumerical("SetMaxVelocityStep", device_id, ATTR_MAXIMUM_VELOCITY,
                                                           maximum_velocity,
                                                           trigger_log=False,
                                                           negative_tolerance=mv.negative_tolerance(),
                                                           positive_tolerance=mv.positive_tolerance())
            position_step_0 = StepDeviceAttributeNumerical("SetInitPositionStep", device_id, ATTR_POSITION,
                                                           scanner_input.start_point,
                                                           trigger_log=False,
                                                           negative_tolerance=scanner_input.step_attributes[
                                                               "negative_tolerance"],
                                                           positive_tolerance=scanner_input.step_attributes[
                                                               "positive_tolerance"])
            macro.add(velocity_step_0)
            macro.add(position_step_0)
        velocity_step_1 = StepDeviceAttributeNumerical("FinalVelocityStep", device_id, ATTR_MAXIMUM_VELOCITY,
                                                       final_velocity,
                                                       trigger_log=False, negative_tolerance=mv.negative_tolerance(),
                                                       positive_tolerance=mv.positive_tolerance())
        macro.add(velocity_step_1)
        macro.add(scanner_input)
        # velocity_step_2 = StepDeviceAttributeNumerical("PreviousVelocityStep", device_id, ATTR_MAXIMUM_VELOCITY,
        #                                              current_velocity,
        #                                              trigger_log=False, negative_tolerance=mv.negative_tolerance(),
        #                                              positive_tolerance=mv.positive_tolerance())
        # macro.add(velocity_step_2)
        return macro



from PyQt5.QtCore import pyqtSignal, pyqtSlot
from PyQt5.QtWidgets import QWidget

import kamzik3
from kamzik3.constants import *
from kamzik3.gui.deviceDisplayAttributeWidget import DeviceDisplayAttributeWidget
from kamzik3.gui.templates.deviceListTemplate import Ui_Form
from kamzik3.snippets.snippetsWidgets import clear_layout
from kamzik3.snippets.snippetsYaml import YamlSerializable


class DeviceListWidget(Ui_Form, QWidget, YamlSerializable):

    def __init__(self, columns, title=None, config=None, parent=None):
        super(DeviceListWidget, self).__init__(parent)
        self.config = config
        self.title = title
        self.widgets = []
        self.columns = columns
        self.setupUi(self)

    def yaml_mapping(self):
        mapping = super().yaml_mapping()
        del mapping["parent"]
        return mapping

    def setupUi(self, Form):
        super(DeviceListWidget, self).setupUi(Form)

        if self.title is not None:
            self.label_device_id.setText(self.title)

        layout = self.widget_content.layout()
        clear_layout(self, layout)

    def add_widget(self, widget):
        layout = self.widget_content.layout()
        row = layout.rowCount() - 1
        new_row = True
        for column in range(self.columns):
            if not layout.itemAtPosition(row, column):
                new_row = False
                break
        if new_row:
            row += 1
            column = 0
        self.widgets.append(widget)
        layout.addWidget(widget, row, column)

    def filter_devices(self):
        filter_string = str(self.input_command.text())
        for widget in self.widgets:
            if widget.device.device_id.find(filter_string) != -1:
                widget.show()
            else:
                widget.hide()

    def remove_widget(self, widget):
        widget.close()
        self.widgets.remove(widget)


class DeviceListWidgetSession(DeviceListWidget):
    sigDeviceCountChanged = pyqtSignal("PyQt_PyObject", "PyQt_PyObject")

    def __init__(self, columns, title=None, parent=None):
        super(DeviceListWidgetSession, self).__init__(columns, title, parent)
        self.sigDeviceCountChanged.connect(self.device_count_changed)
        kamzik3.session.attach_attribute_callback(ATTR_DEVICES_COUNT, self.sigDeviceCountChanged.emit)

    @pyqtSlot("PyQt_PyObject", "PyQt_PyObject")
    def device_count_changed(self, key, value):
        """
        This slot is called when device counter in current session has been changed.
        Get thru all widgets and remove or add new / old devices from list.
        :param key:
        :param value:
        :return:
        """
        current_devices = list(kamzik3.session.devices.keys())
        for widget in self.widgets[:]:
            if widget.device.device_id not in current_devices:
                self.remove_widget(widget)
            else:
                current_devices.remove(widget.device.device_id)
        for device_id in current_devices:
            self.add_widget(DeviceDisplayAttributeWidget(device=device_id, attribute=ATTR_LATENCY,
                                                         config={"max_update_rate": 400}))

    def close(self):
        kamzik3.session.detach_attribute_callback(ATTR_DEVICES_COUNT, self.sigDeviceCountChanged.emit)
        return super(DeviceListWidgetSession, self).close()



import json
import logging
from collections import deque
from threading import Thread, Event
from time import perf_counter_ns

import zmq

import kamzik3
from kamzik3 import SEP, CommandFormatException, DeviceServerError, DeviceError
from kamzik3.constants import *
from kamzik3.snippets.snippetDataAccess import rgetattr
from kamzik3.snippets.snippetsJson import JsonPintHook
from kamzik3.snippets.snippetsTimer import PreciseCallbackTimer
from kamzik3.snippets.snippetsYaml import YamlSerializable


class DeviceServer(YamlSerializable):

    def __init__(self, host, server_port, publisher_port):
        self.logger = logging.getLogger(u"Console.DeviceServer@{}:{}".format(host, server_port))
        self.host = host
        self.server_port = server_port
        self.publisher_port = publisher_port
        self.close_event = Event()
        self.messages_queue = deque()
        self.publisher_events = {}
        self.devices = {}
        self._init_publisher()
        self._init_server()
        kamzik3.session.set_device_server(self)

    def _init_server(self):
        """
        Initialize connection to Device server as a client.
        Connection is done via ZMQ REP socket.
        :return: None
        """
        self.logger.info(u"Initializing server thread")
        self.server_socket = zmq.Context.instance().socket(zmq.REP)
        # Set receive timeout to 1000 ms
        self.server_socket.setsockopt(zmq.RCVTIMEO, 1000)
        # Don't let socket linger after close
        # This is VERY IMPORTANT not to set to anything but 0
        self.server_socket.setsockopt(zmq.LINGER, 0)
        self.server_socket.bind("tcp://{}:{}".format(self.host, self.server_port))
        self.server_thread = Thread(target=self._server_thread)
        self.server_thread.start()

    def _server_thread(self):
        while not self.close_event.isSet():
            if self.server_socket.poll(1e3):
                try:
                    data = self.server_socket.recv().decode().split(SEP)
                    device_id, instruction = data[0], data[1]

                    # Execute command directly on devices
                    if instruction == COMMAND:
                        command, with_token = data[2], int(data[3])
                        token = self.get_device(device_id).command(command, with_token=int(with_token))
                        status, response = RESPONSE_OK, len(data[3])

                    # Get devices attribute
                    elif instruction == GET:
                        attribute = json.loads(data[2], object_hook=JsonPintHook)
                        response = self.get_device(device_id)._get(attribute)
                        response = json.dumps(response, ensure_ascii=True)
                        status, token = RESPONSE_OK, 0

                    # Set devices attribute
                    elif instruction == SET:
                        attribute, attribute_value = json.loads(data[2], object_hook=JsonPintHook)
                        set_token = self.get_device(device_id).set_attribute(attribute, attribute_value)
                        status, token, response = RESPONSE_OK, set_token, len(data[2])

                    # Execute devices method with attributes
                    elif instruction == METHOD:
                        method, attributes = data[2], json.loads(data[3], object_hook=JsonPintHook)
                        device = self.get_device(device_id)
                        response = rgetattr(device, method)(**attributes)
                        response = json.dumps(response, ensure_ascii=True)
                        status, token = RESPONSE_OK, 0

                    # Poll devices for activity
                    elif instruction == POLL:
                        status, token, response = RESPONSE_OK, 0, 1

                    # Init devices
                    elif instruction == INIT:
                        device = self.get_device(device_id)

                        if device is None:
                            raise DeviceServerError(
                                "Device {} is not registered on the server or publisher is not ready".format(device_id))
                        elif self.publisher_port is None:
                            raise DeviceServerError("Server publisher is not ready")
                        elif not device.in_statuses(READY_DEVICE_STATUSES):
                            raise DeviceServerError("Device is not ready")
                        else:
                            response = json.dumps(
                                (self.publisher_port, device.attributes, device.attributes_sharing_map,
                                 device.exposed_methods, device.qualified_name),
                                ensure_ascii=True)
                            status, token = RESPONSE_OK, 0

                    # None of above, request not implemented
                    else:
                        status, token, response = RESPONSE_ERROR, 0, len(data[3])

                    self.server_socket.send(SEP.join((str(status), str(token), str(response))).encode())
                except CommandFormatException:
                    status, token, response = RESPONSE_ERROR, 0, u"Command format error"
                    self.server_socket.send(SEP.join((str(status), str(token), str(response))).encode())
                except DeviceServerError:
                    status, token, response = RESPONSE_ERROR, 0, u"Device server error"
                    self.server_socket.send(SEP.join((str(status), str(token), str(response))).encode())
                except DeviceError as e:
                    status, token, response = RESPONSE_ERROR, 0, u"Command error: {}".format(e)
                    self.server_socket.send(SEP.join((str(status), str(token), str(response))).encode())
                except (AttributeError, KeyError) as e:
                    status, token, response = RESPONSE_ERROR, 0, u"Attribute error: {}".format(e)
                    self.server_socket.send(SEP.join((str(status), str(token), str(response))).encode())

        self.handle_closing_server()

    def _init_publisher(self):
        self.logger.info(u"Initializing publisher thread")
        self.publisherSocket = zmq.Context.instance().socket(zmq.PUB)
        self.publisherSocket.bind("tcp://{}:{}".format(self.host, self.publisher_port))
        self.publisher_timer = PreciseCallbackTimer(5, self._publish_messages)
        self.publisher_timer.start()

    def _publish_messages(self):
        publish_time = perf_counter_ns()
        while self.messages_queue:
            header, message, min_timeout = self.messages_queue.popleft()
            last_message_event = self.publisher_events.get(header, None)
            if (last_message_event is None) or (publish_time - last_message_event[0]) >= min_timeout:
                self._publish_message(header, message)
                self.publisher_events[header] = [publish_time, [header, None, min_timeout]]
            else:
                last_message_event[1][1] = message

        for event in list(self.publisher_events.values()):
            if (publish_time - event[0]) >= event[1][2]:
                if event[1][1] is not None:
                    self._publish_message(event[1][0], event[1][1])
                    self.publisher_events[event[1][0]] = [publish_time, [event[1][0], None, event[1][2]]]
                else:
                    del self.publisher_events[event[1][0]]

    def _publish_message(self, header, message):
        try:
            out_message = header + SEP + json.dumps(message, ensure_ascii=True)
            self.publisherSocket.send(out_message.encode())
        except TypeError:
            print(message)

    def add_device(self, device):
        self.logger.info(u"Adding {} into server".format(device.device_id))
        if device.device_id not in self.devices:
            self.devices[device.device_id] = device
            device.set_device_server(self)
            return True
        else:
            return False

    def remove_device(self, device):
        self.logger.info(u"Removing {} from server".format(device.device_id))
        try:
            del self.devices[device.device_id]
            device.clear_device_server()
            return True
        except ValueError:
            return False

    def get_device(self, device_id):
        return self.devices.get(device_id, None)

    def push_message(self, header, message, token=None, min_timeout=100e6):
        try:
            if token is not None:
                header += u"." + str(token)
            self.messages_queue.append((header, message, min_timeout))
        except TypeError:
            self.logger.exception(u"Error pushing message to client")

    def handle_closing_server(self):
        if self.server_socket:
            self.server_socket.close()

    def close(self):
        self.logger.info(u"Closing server")
        self.publisher_timer.stop()
        self.close_event.set()
        if self.server_thread is not None:
            self.server_thread.join(1)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'sessionWindowTemplate.ui'
#
# Created by: PyQt5 UI code generator 5.10.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(166, 76)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/icons/kamzik.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        MainWindow.setWindowIcon(icon)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout.setObjectName("gridLayout")
        self.label_header = QtWidgets.QLabel(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_header.sizePolicy().hasHeightForWidth())
        self.label_header.setSizePolicy(sizePolicy)
        self.label_header.setStyleSheet("font-weight:bold;background:#f6b442;padding:5")
        self.label_header.setObjectName("label_header")
        self.gridLayout.addWidget(self.label_header, 0, 0, 1, 1)
        self.layout_tool_buttons = QtWidgets.QGridLayout()
        self.layout_tool_buttons.setObjectName("layout_tool_buttons")
        self.gridLayout.addLayout(self.layout_tool_buttons, 1, 0, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem, 2, 0, 1, 1)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 166, 21))
        self.menubar.setObjectName("menubar")
        self.menuApplication = QtWidgets.QMenu(self.menubar)
        self.menuApplication.setObjectName("menuApplication")
        MainWindow.setMenuBar(self.menubar)
        self.actionExit = QtWidgets.QAction(MainWindow)
        self.actionExit.setObjectName("actionExit")
        self.menuApplication.addAction(self.actionExit)
        self.menubar.addAction(self.menuApplication.menuAction())

        self.retranslateUi(MainWindow)
        self.actionExit.triggered.connect(MainWindow.close)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "MainWindow"))
        self.label_header.setText(_translate("MainWindow", "Session tools"))
        self.menuApplication.setTitle(_translate("MainWindow", "Application"))
        self.actionExit.setText(_translate("MainWindow", "Exit"))

from kamzik3 import resource_kamzik3_rc



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'deviceListTemplate.ui'
#
# Created by: PyQt5 UI code generator 5.10.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(567, 348)
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setObjectName("gridLayout")
        self.label_device_id = QtWidgets.QLabel(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_device_id.sizePolicy().hasHeightForWidth())
        self.label_device_id.setSizePolicy(sizePolicy)
        self.label_device_id.setMinimumSize(QtCore.QSize(0, 35))
        font = QtGui.QFont()
        font.setPointSize(23)
        font.setBold(True)
        font.setWeight(75)
        self.label_device_id.setFont(font)
        self.label_device_id.setStyleSheet("QLabel{background:#68d532;}")
        self.label_device_id.setAlignment(QtCore.Qt.AlignCenter)
        self.label_device_id.setObjectName("label_device_id")
        self.gridLayout.addWidget(self.label_device_id, 0, 0, 1, 1)
        self.widget_scroll_content = QtWidgets.QScrollArea(Form)
        self.widget_scroll_content.setStyleSheet("")
        self.widget_scroll_content.setWidgetResizable(True)
        self.widget_scroll_content.setObjectName("widget_scroll_content")
        self.widget_content = QtWidgets.QWidget()
        self.widget_content.setGeometry(QtCore.QRect(0, 0, 547, 253))
        self.widget_content.setStyleSheet("QWidget#widget_content {background:#fff}")
        self.widget_content.setObjectName("widget_content")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.widget_content)
        self.gridLayout_3.setContentsMargins(0, 0, 0, 0)
        self.gridLayout_3.setSpacing(0)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.widget_scroll_content.setWidget(self.widget_content)
        self.gridLayout.addWidget(self.widget_scroll_content, 3, 0, 1, 1)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.input_command = QtWidgets.QLineEdit(Form)
        self.input_command.setInputMask("")
        self.input_command.setObjectName("input_command")
        self.horizontalLayout.addWidget(self.input_command)
        self.pushButton = QtWidgets.QPushButton(Form)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/icons/search.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.pushButton.setIcon(icon)
        self.pushButton.setObjectName("pushButton")
        self.horizontalLayout.addWidget(self.pushButton)
        self.gridLayout.addLayout(self.horizontalLayout, 2, 0, 1, 1)

        self.retranslateUi(Form)
        self.pushButton.clicked.connect(Form.filter_devices)
        self.input_command.returnPressed.connect(self.pushButton.click)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.label_device_id.setText(_translate("Form", "Device list"))
        self.input_command.setPlaceholderText(_translate("Form", "Filter devices by name"))
        self.pushButton.setText(_translate("Form", "Filter"))

from kamzik3 import resource_kamzik3_rc



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'macroStepTemplate.ui'
#
# Created by: PyQt5 UI code generator 5.13.0
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(320, 189)
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setObjectName("gridLayout")
        self.combobox_step = QtWidgets.QComboBox(Form)
        self.combobox_step.setSizeAdjustPolicy(QtWidgets.QComboBox.AdjustToContents)
        self.combobox_step.setObjectName("combobox_step")
        self.gridLayout.addWidget(self.combobox_step, 3, 1, 1, 1)
        self.widget_settings = QtWidgets.QWidget(Form)
        self.widget_settings.setObjectName("widget_settings")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.widget_settings)
        self.gridLayout_2.setContentsMargins(0, 0, 0, 0)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.checkbox_defaults = QtWidgets.QCheckBox(self.widget_settings)
        self.checkbox_defaults.setChecked(False)
        self.checkbox_defaults.setObjectName("checkbox_defaults")
        self.gridLayout_2.addWidget(self.checkbox_defaults, 0, 1, 1, 1)
        self.checkbox_log = QtWidgets.QCheckBox(self.widget_settings)
        self.checkbox_log.setChecked(True)
        self.checkbox_log.setObjectName("checkbox_log")
        self.gridLayout_2.addWidget(self.checkbox_log, 0, 0, 1, 1)
        self.gridLayout.addWidget(self.widget_settings, 1, 1, 1, 1)
        self.label_scanner = QtWidgets.QLabel(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_scanner.sizePolicy().hasHeightForWidth())
        self.label_scanner.setSizePolicy(sizePolicy)
        self.label_scanner.setMaximumSize(QtCore.QSize(110, 16777215))
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.label_scanner.setFont(font)
        self.label_scanner.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.label_scanner.setObjectName("label_scanner")
        self.gridLayout.addWidget(self.label_scanner, 6, 0, 1, 1)
        self.label_method = QtWidgets.QLabel(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_method.sizePolicy().hasHeightForWidth())
        self.label_method.setSizePolicy(sizePolicy)
        self.label_method.setMaximumSize(QtCore.QSize(110, 16777215))
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.label_method.setFont(font)
        self.label_method.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.label_method.setObjectName("label_method")
        self.gridLayout.addWidget(self.label_method, 5, 0, 1, 1)
        self.combobox_device = QtWidgets.QComboBox(Form)
        self.combobox_device.setSizeAdjustPolicy(QtWidgets.QComboBox.AdjustToContents)
        self.combobox_device.setObjectName("combobox_device")
        self.gridLayout.addWidget(self.combobox_device, 2, 1, 1, 1)
        self.label_step = QtWidgets.QLabel(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_step.sizePolicy().hasHeightForWidth())
        self.label_step.setSizePolicy(sizePolicy)
        self.label_step.setMaximumSize(QtCore.QSize(110, 16777215))
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.label_step.setFont(font)
        self.label_step.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.label_step.setObjectName("label_step")
        self.gridLayout.addWidget(self.label_step, 3, 0, 1, 1)
        self.label_attribute = QtWidgets.QLabel(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_attribute.sizePolicy().hasHeightForWidth())
        self.label_attribute.setSizePolicy(sizePolicy)
        self.label_attribute.setMaximumSize(QtCore.QSize(110, 16777215))
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.label_attribute.setFont(font)
        self.label_attribute.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.label_attribute.setObjectName("label_attribute")
        self.gridLayout.addWidget(self.label_attribute, 4, 0, 1, 1)
        self.label_settings = QtWidgets.QLabel(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_settings.sizePolicy().hasHeightForWidth())
        self.label_settings.setSizePolicy(sizePolicy)
        self.label_settings.setMaximumSize(QtCore.QSize(110, 16777215))
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.label_settings.setFont(font)
        self.label_settings.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTop|QtCore.Qt.AlignTrailing)
        self.label_settings.setObjectName("label_settings")
        self.gridLayout.addWidget(self.label_settings, 1, 0, 1, 1)
        self.label_device = QtWidgets.QLabel(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_device.sizePolicy().hasHeightForWidth())
        self.label_device.setSizePolicy(sizePolicy)
        self.label_device.setMaximumSize(QtCore.QSize(110, 16777215))
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.label_device.setFont(font)
        self.label_device.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.label_device.setObjectName("label_device")
        self.gridLayout.addWidget(self.label_device, 2, 0, 1, 1)
        self.combobox_attribute = QtWidgets.QComboBox(Form)
        self.combobox_attribute.setSizeAdjustPolicy(QtWidgets.QComboBox.AdjustToContents)
        self.combobox_attribute.setObjectName("combobox_attribute")
        self.gridLayout.addWidget(self.combobox_attribute, 4, 1, 1, 1)
        self.combobox_method = QtWidgets.QComboBox(Form)
        self.combobox_method.setSizeAdjustPolicy(QtWidgets.QComboBox.AdjustToContents)
        self.combobox_method.setObjectName("combobox_method")
        self.gridLayout.addWidget(self.combobox_method, 5, 1, 1, 1)
        self.combobox_scanner = QtWidgets.QComboBox(Form)
        self.combobox_scanner.setSizeAdjustPolicy(QtWidgets.QComboBox.AdjustToContents)
        self.combobox_scanner.setObjectName("combobox_scanner")
        self.gridLayout.addWidget(self.combobox_scanner, 6, 1, 1, 1)
        self.button_settings = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_settings.sizePolicy().hasHeightForWidth())
        self.button_settings.setSizePolicy(sizePolicy)
        self.button_settings.setText("")
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/icons/settings.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_settings.setIcon(icon)
        self.button_settings.setIconSize(QtCore.QSize(19, 19))
        self.button_settings.setObjectName("button_settings")
        self.gridLayout.addWidget(self.button_settings, 2, 2, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(1, 20, QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem, 2, 3, 1, 1)
        self.frame_scan_select = QtWidgets.QFrame(Form)
        self.frame_scan_select.setStyleSheet("QFrame {background:#a9c7db}")
        self.frame_scan_select.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.frame_scan_select.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_scan_select.setObjectName("frame_scan_select")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.frame_scan_select)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setSpacing(0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.label_axis_settings = QtWidgets.QLabel(self.frame_scan_select)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_axis_settings.sizePolicy().hasHeightForWidth())
        self.label_axis_settings.setSizePolicy(sizePolicy)
        self.label_axis_settings.setStyleSheet("QLabel {font-weight:bold;padding:5}")
        self.label_axis_settings.setObjectName("label_axis_settings")
        self.horizontalLayout.addWidget(self.label_axis_settings)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.gridLayout.addWidget(self.frame_scan_select, 0, 0, 1, 4)
        spacerItem2 = QtWidgets.QSpacerItem(0, 0, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem2, 7, 0, 1, 3)

        self.retranslateUi(Form)
        self.button_settings.clicked.connect(Form.show_debug_widget)
        self.combobox_device.currentTextChanged['QString'].connect(Form.device_selected)
        self.combobox_step.currentIndexChanged['QString'].connect(Form.step_selected)
        self.combobox_attribute.currentIndexChanged['QString'].connect(Form.attribute_selected)
        self.combobox_method.currentIndexChanged['QString'].connect(Form.method_selected)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.checkbox_defaults.setToolTip(_translate("Form", "Check if You want to set default values for this scan"))
        self.checkbox_defaults.setText(_translate("Form", "Set defaults"))
        self.checkbox_log.setToolTip(_translate("Form", "Check if You want to insert line into log when scan finishes one step"))
        self.checkbox_log.setText(_translate("Form", "Trigger log"))
        self.label_scanner.setText(_translate("Form", "Scanner:"))
        self.label_method.setText(_translate("Form", "Method:"))
        self.label_step.setText(_translate("Form", "Step:"))
        self.label_attribute.setText(_translate("Form", "Attribute:"))
        self.label_settings.setText(_translate("Form", "Settings:"))
        self.label_device.setText(_translate("Form", "Device:"))
        self.button_settings.setToolTip(_translate("Form", "Show device debug widget"))
        self.label_axis_settings.setText(_translate("Form", "Step settings"))
from kamzik3 import resource_kamzik3_rc



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'deviceGroupedListTemplate.ui'
#
# Created by: PyQt5 UI code generator 5.10.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(415, 264)
        MainWindow.setStyleSheet("QWidget#centralwidget {background:#dde9ff}")
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setStyleSheet("centralwidget {background:#dde9ff}")
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout.setContentsMargins(6, 6, 6, 6)
        self.gridLayout.setObjectName("gridLayout")
        self.layout_top_controls = QtWidgets.QWidget(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.layout_top_controls.sizePolicy().hasHeightForWidth())
        self.layout_top_controls.setSizePolicy(sizePolicy)
        self.layout_top_controls.setMinimumSize(QtCore.QSize(0, 200))
        self.layout_top_controls.setMaximumSize(QtCore.QSize(16777215, 200))
        self.layout_top_controls.setStyleSheet("QWidget#layout_top_controls {\n"
"    border-bottom: 2px solid #d1d1d1;\n"
"    background: #e3e3e3;\n"
"}")
        self.layout_top_controls.setObjectName("layout_top_controls")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.layout_top_controls)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setSpacing(0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.gridLayout.addWidget(self.layout_top_controls, 0, 0, 1, 1)
        self.tab_device_groups = QtWidgets.QTabWidget(self.centralwidget)
        self.tab_device_groups.setStyleSheet("QTabBar::tab:selected {\n"
"    font-weight: bold;\n"
"    color: #003fb4;\n"
"}")
        self.tab_device_groups.setTabShape(QtWidgets.QTabWidget.Rounded)
        self.tab_device_groups.setObjectName("tab_device_groups")
        self.gridLayout.addWidget(self.tab_device_groups, 1, 0, 1, 1)
        self.layout_bottom_controls = QtWidgets.QHBoxLayout()
        self.layout_bottom_controls.setContentsMargins(0, -1, 0, -1)
        self.layout_bottom_controls.setObjectName("layout_bottom_controls")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.layout_bottom_controls.addItem(spacerItem)
        self.button_stop_all_devices = QtWidgets.QPushButton(self.centralwidget)
        self.button_stop_all_devices.setMinimumSize(QtCore.QSize(0, 0))
        self.button_stop_all_devices.setCursor(QtGui.QCursor(QtCore.Qt.OpenHandCursor))
        self.button_stop_all_devices.setFocusPolicy(QtCore.Qt.NoFocus)
        self.button_stop_all_devices.setStyleSheet("QPushButton { background-color: #ff4e00 }\\\\nQPushButton:pressed { background-color: #ff4e00 }\\\\nQPushButton:!enabled{background:silver;}QPushButton:checked{background:red;}\\nQPushButton:disabled{ background:silver;color:grey }")
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/icons/stop.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_stop_all_devices.setIcon(icon)
        self.button_stop_all_devices.setIconSize(QtCore.QSize(20, 24))
        self.button_stop_all_devices.setAutoExclusive(False)
        self.button_stop_all_devices.setObjectName("button_stop_all_devices")
        self.layout_bottom_controls.addWidget(self.button_stop_all_devices)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.layout_bottom_controls.addItem(spacerItem1)
        self.button_unlock_controls = QtWidgets.QPushButton(self.centralwidget)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.button_unlock_controls.setFont(font)
        self.button_unlock_controls.setCursor(QtGui.QCursor(QtCore.Qt.OpenHandCursor))
        self.button_unlock_controls.setStyleSheet("QPushButton { background-color: orange }\\\\nQPushButton:pressed { background-color: #ff4e00 }\\\\nQPushButton:!enabled{background:silver;}QPushButton:checked{background:red;}\\nQPushButton:disabled{ background:silver;color:grey }")
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/icons/icons/unlock.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_unlock_controls.setIcon(icon1)
        self.button_unlock_controls.setIconSize(QtCore.QSize(20, 24))
        self.button_unlock_controls.setObjectName("button_unlock_controls")
        self.layout_bottom_controls.addWidget(self.button_unlock_controls)
        self.button_lock_controls = QtWidgets.QPushButton(self.centralwidget)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.button_lock_controls.setFont(font)
        self.button_lock_controls.setCursor(QtGui.QCursor(QtCore.Qt.OpenHandCursor))
        self.button_lock_controls.setStyleSheet("QPushButton { background-color: orange }\\\\nQPushButton:pressed { background-color: #ff4e00 }\\\\nQPushButton:!enabled{background:silver;}QPushButton:checked{background:red;}\\nQPushButton:disabled{ background:silver;color:grey }")
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/icons/icons/lock.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_lock_controls.setIcon(icon2)
        self.button_lock_controls.setIconSize(QtCore.QSize(20, 24))
        self.button_lock_controls.setObjectName("button_lock_controls")
        self.layout_bottom_controls.addWidget(self.button_lock_controls)
        self.gridLayout.addLayout(self.layout_bottom_controls, 2, 0, 1, 1)
        MainWindow.setCentralWidget(self.centralwidget)

        self.retranslateUi(MainWindow)
        self.button_lock_controls.clicked.connect(MainWindow.lock_widget)
        self.button_stop_all_devices.clicked.connect(MainWindow.stop_all_devices)
        self.button_unlock_controls.clicked.connect(MainWindow.unlock_widget)
        self.tab_device_groups.currentChanged['int'].connect(MainWindow.tab_changed)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "MainWindow"))
        self.button_stop_all_devices.setText(_translate("MainWindow", " Stop all devices (ESC) "))
        self.button_unlock_controls.setToolTip(_translate("MainWindow", "Click to unlock device controls"))
        self.button_unlock_controls.setText(_translate("MainWindow", "Unlock controls "))
        self.button_lock_controls.setToolTip(_translate("MainWindow", "Click to lock device controls"))
        self.button_lock_controls.setText(_translate("MainWindow", " Lock controls "))

from kamzik3 import resource_kamzik3_rc



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="kamzik3",
    version="0.5.7",
    author="Martin Domaracky",
    author_email="martin.domaracky@desy.de",
    description="Device controlling framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://stash.desy.de/projects/KAM/repos/kamzik3",
    packages=setuptools.find_packages(),
    package_data={'kamzik3': ['*.yml'], 'kamzik3.examples': ['*.yml', "*.att"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Natural Language :: English",
        "Intended Audience :: Science/Research",
    ],
    install_requires=[
        "numpy>=1.16",
        "pyzmq>=18.0",
        "pint>=0.10.1",
        "bidict>=0.18",
        "pyqt5>=5.13",
        "pyqtgraph>=0.10",
        "pyserial>=3.4",
        "oyaml>=0.9",
        "psutil>=5.6.0",
    ],
    extras_requires={
        "pytango": [],
        "pyopengl": [],
        "sysutil": [],
        "pydaqmx": [],
        "pypiwin32": [],
    },
    python_requires='>=3.6',
)



import logging
import os
import time
from collections import OrderedDict

import numpy as np
import psutil

import kamzik3
from kamzik3 import DeviceUnknownError, DeviceError
from kamzik3.constants import *
from kamzik3.devices.attributeLogger import AttributeLogger, ATTR_LOGGING
from kamzik3.devices.device import Device
from kamzik3.devices.deviceClient import DeviceClient
from kamzik3.devices.general.measurementGroup import MeasurementGroup
from kamzik3.snippets.snippetLogging import set_rotating_file_handler, set_sys_exception_handler, get_console_handler
from kamzik3.snippets.snippetsControlLoops import control_asyncore_loop, control_device_poller, control_port_read_loop, \
    control_device_connection_poller
from kamzik3.snippets.snippetsTimer import CallbackTimer


class DeviceSession(Device):
    LOG_DEVICE_OUTPUT_FILE = u"devices.log"
    LOG_CONSOLE_OUTPUT_FILE = u"console.log"
    LOG_ATTRIBUTES_OUTPUT_FILE = u"attributes.log"
    LOG_GUI_OUTPUT_FILE = u"gui.log"
    LOG_MACRO_OUTPUT_FILE = u"macro.log"
    log_level = logging.INFO

    def __init__(self, device_id=None, config=None):
        kamzik3.session = self
        self.devices = OrderedDict()
        self.measurement_groups = {}
        self.polling_timer = CallbackTimer(1000, self.measure_cpu_ram)
        self.os_process = psutil.Process(os.getpid())
        super(DeviceSession, self).__init__(device_id, config)
        self.connect()

    def _init_attributes(self):
        super(DeviceSession, self)._init_attributes()
        self.create_attribute(ATTR_LOG_DIRECTORY, set_function=self.set_log_directory, readonly=True)
        self.create_attribute(ATTR_RESOURCE_DIRECTORY, set_function=self.set_resource_directory, readonly=True)
        self.create_attribute(ATTR_DEVICES_COUNT, default_value=0, default_type=np.uint64, readonly=True)
        self.create_attribute(ATTR_ALLOW_CONSOLE_LOG, default_value=True, default_type=np.bool)
        self.create_attribute(ATTR_ALLOW_GUI_LOG, default_value=True, default_type=np.bool)
        self.create_attribute(ATTR_ALLOW_DEVICE_LOG, default_value=True, default_type=np.bool)
        self.create_attribute(ATTR_ALLOW_ATTRIBUTE_LOG, default_value=True, default_type=np.bool)
        self.create_attribute(ATTR_ALLOW_MACRO_LOG, default_value=True, default_type=np.bool)
        self.create_attribute(ATTR_LOOPS_RUNNING, default_value=False, default_type=np.bool, readonly=True)
        self.create_attribute(ATTR_PROCESS_ID, default_value=os.getpid(), readonly=True)
        self.create_attribute(ATTR_CPU_USAGE, default_value=0, default_type=np.float, readonly=True, unit="%",
                              decimals=4)
        self.create_attribute(ATTR_MEMORY_USAGE, default_value=0, default_type=np.float, readonly=True, decimals=3,
                              unit="MB", factor=10e-7)
        self.create_attribute(ATTR_AVAILABLE_MEMORY, default_value=0, default_type=np.float,
                              readonly=True, unit="MB", factor=10e-7, decimals=3)
        self.set_raw_value(ATTR_AVAILABLE_MEMORY, psutil.virtual_memory().total)

    def add_measurement_group(self, measurement_group):
        assert isinstance(measurement_group, MeasurementGroup)
        self.measurement_groups[measurement_group.group_id] = measurement_group

    def set_session(self, session):
        self.devices[self.device_id] = self
        self.session = self
        self.set_value(ATTR_DEVICES_COUNT, len(self.devices))

    def measure_cpu_ram(self):
        self.set_raw_value(ATTR_CPU_USAGE, self.os_process.cpu_percent())
        self.set_raw_value(ATTR_MEMORY_USAGE, self.os_process.memory_info().rss)

    def handle_configuration(self):
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()

            if self.get_value(ATTR_ALLOW_CONSOLE_LOG):
                self.set_console_log_directory(u"Console")
            if self.get_value(ATTR_ALLOW_GUI_LOG):
                self.set_gui_log_directory(u"Gui")
            if self.get_value(ATTR_ALLOW_DEVICE_LOG):
                self.set_device_log_directory(u"Device")
            if self.get_value(ATTR_ALLOW_ATTRIBUTE_LOG):
                self.set_attribute_log_directory(u"Attributes")
                self._config_attributes_log()
            if self.get_value(ATTR_ALLOW_MACRO_LOG):
                self.set_macro_log_directory(u"Macro")

            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        _finish_configuration()

    def _config_attributes_log(self):
        if self.config:
            logger = self.get_device("AttributeLogger")
            for device_id, attribute in self.config.get("logged_attributes", {}).items():
                if isinstance(attribute, list):
                    for sub_attribute in attribute:
                        logger.add_logged_attribute(device_id, sub_attribute)
                else:
                    logger.add_logged_attribute(device_id, attribute)
            logger.set_attribute((ATTR_LOGGING, VALUE), True)

    def start_polling(self):
        self.polling_timer.start()

    def stop_polling(self):
        self.polling_timer.stop()

    def command(self, command, callback=None, with_token=False, returning=True):
        raise DeviceError("Device does not accept any commands.")

    def get_device(self, device_id):
        try:
            return self.devices[device_id]
        except KeyError:
            master_server, master_port = self.config.get("master_server", False), self.config.get("master_port", False)
            if master_server and master_port:
                return DeviceClient(master_server, master_port, device_id)
            raise DeviceUnknownError("Device {} was not found.".format(device_id))

    def change_device_id(self, device_id, new_device_id):
        device = self.get_device(device_id)
        device.device_id = new_device_id
        self.devices[new_device_id] = self.devices.pop(device_id)

    def add_device(self, device):
        assert isinstance(device, Device)

        if device.device_id not in self.devices:
            self.devices[device.device_id] = device
            device.session = self
            self.set_value(ATTR_DEVICES_COUNT, len(self.devices))
            if self.device_server is not None:
                self.device_server.add_device(device)
        else:

            raise DeviceError("Device {} is already registered within session.".format(device.device_id))

    def remove_device(self, device):
        assert isinstance(device, Device)
        if device.device_id in self.devices:
            device.session = None
            del self.devices[device.device_id]
            self.set_value(ATTR_DEVICES_COUNT, len(self.devices))

    def get_devices(self, class_filter=None, attribute_filter=None, method_filter=None):
        """
        Filter devices in current session
        :param class_filter: list of classes
        :param attribute_filter: attribute
        :param method_filter: method name
        :return: list of filtered devices
        """
        filtered_devices = self.devices.copy()
        # Filter over device class
        if class_filter is not None:
            for device_id, device in filtered_devices.copy().items():
                if class_filter in device.qualified_name:
                    del filtered_devices[device_id]
        # Filter over device attributes
        if attribute_filter is not None:
            for device_id, device in filtered_devices.copy().items():
                if device.get_attribute(attribute_filter) is None:
                    del filtered_devices[device_id]
        # Filter over device methods
        if method_filter is not None:
            for device_id, device in filtered_devices.copy().items():
                filtered_out_methods = filter(lambda val: val[0] == method_filter, device.exposed_methods)
                try:
                    next(filtered_out_methods)
                except StopIteration:
                    del filtered_devices[device_id]

        return filtered_devices

    def set_log_directory(self, value):
        self.logger.info(u"Setting log directory to: {}".format(value))
        if not os.path.exists(value):
            self.logger.info(u"Directory {} does not exists, trying to create it.".format(value))
            os.makedirs(value)

    def set_resource_directory(self, value):
        self.logger.info(u"Setting resource directory to: {}".format(value))
        if not os.path.exists(value):
            self.logger.info(u"Directory {} does not exists, trying to create it.".format(value))
            os.makedirs(value)

    def set_console_log_directory(self, value):
        log_directory = self.get_value(ATTR_LOG_DIRECTORY)
        if log_directory is None:
            raise DeviceError(u"Cannot set console log directory. Attribute 'Log directory' is not set.")
        log_directory = os.path.join(log_directory, value)
        self.logger.info(u"Setting console log directory to: {}".format(log_directory))
        if not os.path.exists(log_directory):
            self.logger.info(u"Directory {} does not exists, trying to create it.".format(log_directory))
            os.makedirs(log_directory)

        handler = get_console_handler()
        logger = logging.getLogger(u"Console")
        set_rotating_file_handler(logger, os.path.join(log_directory, self.LOG_CONSOLE_OUTPUT_FILE))
        set_sys_exception_handler(logger)
        logger.setLevel(self.log_level)
        logger.addHandler(handler)

    def set_gui_log_directory(self, value):
        log_directory = self.get_value(ATTR_LOG_DIRECTORY)
        if log_directory is None:
            raise DeviceError(u"Cannot set gui log directory. Attribute 'Log directory' is not set.")
        log_directory = os.path.join(log_directory, value)

        self.logger.info(u"Setting gui log directory to: {}".format(log_directory))
        if not os.path.exists(log_directory):
            self.logger.info(u"Directory {} does not exists, trying to create it.".format(log_directory))
            os.makedirs(log_directory)

        handler = get_console_handler()
        logger = logging.getLogger(u"Gui")
        set_rotating_file_handler(logger, os.path.join(log_directory, self.LOG_GUI_OUTPUT_FILE))
        logger.setLevel(logging.DEBUG)
        logger.addHandler(handler)

    def set_attribute_log_directory(self, value):
        log_directory = self.get_value(ATTR_LOG_DIRECTORY)
        if log_directory is None:
            raise DeviceError(u"Cannot set device log directory. Attribute 'Log directory' is not set.")
        log_directory = os.path.join(log_directory, value)

        self.logger.info(u"Setting device log directory to: {}".format(log_directory))
        if not os.path.exists(log_directory):
            self.logger.info(u"Directory {} does not exists, trying to create it.".format(log_directory))
            os.makedirs(log_directory)

        logger = AttributeLogger(os.path.join(log_directory, self.LOG_ATTRIBUTES_OUTPUT_FILE),
                                 device_id="AttributeLogger")
        self.attributes.update({
            "Attribute logger": logger.attributes
        })

    def set_device_log_directory(self, value):
        log_directory = self.get_value(ATTR_LOG_DIRECTORY)
        if log_directory is None:
            raise DeviceError(u"Cannot set device log directory. Attribute 'Log directory' is not set.")
        log_directory = os.path.join(log_directory, value)

        self.logger.info(u"Setting device log directory to: {}".format(log_directory))
        if not os.path.exists(log_directory):
            self.logger.info(u"Directory {} does not exists, trying to create it.".format(log_directory))
            os.makedirs(log_directory)

        handler = get_console_handler()
        logger = logging.getLogger(u"Device")
        set_rotating_file_handler(logger, os.path.join(log_directory, self.LOG_DEVICE_OUTPUT_FILE))
        logger.setLevel(self.log_level)
        logger.addHandler(handler)

    def set_macro_log_directory(self, value):
        log_directory = self.get_value(ATTR_LOG_DIRECTORY)
        if log_directory is None:
            raise DeviceError(u"Cannot set gui log directory. Attribute 'Log directory' is not set.")
        log_directory = os.path.join(log_directory, value)

        self.logger.info(u"Setting gui log directory to: {}".format(log_directory))
        if not os.path.exists(log_directory):
            self.logger.info(u"Directory {} does not exists, trying to create it.".format(log_directory))
            os.makedirs(log_directory)

        handler = get_console_handler()
        logger = logging.getLogger(u"Macro")
        set_rotating_file_handler(logger, os.path.join(log_directory, self.LOG_MACRO_OUTPUT_FILE))
        logger.setLevel(logging.DEBUG)
        logger.addHandler(handler)

    def start_control_loops(self):
        if not self.get_value(ATTR_LOOPS_RUNNING):
            self.start_asyncore_loop()
            self.start_device_connection_poller()
            self.start_device_poller_loop()
            self.start_port_read_loop()
            self.set_value(ATTR_LOOPS_RUNNING, True)

    def stop_control_loops(self):
        if self.get_value(ATTR_LOOPS_RUNNING):
            self.stop_asyncore_loop()
            self.stop_port_read_loop()
            self.stop_device_poller_loop()
            self.stop_device_connection_poller()
            self.set_value(ATTR_LOOPS_RUNNING, False)

        self.logger.info("Closing all control loops")
        # Here we try to join all control loops giving them 1 second timeout
        control_asyncore_loop.join(1)
        control_device_poller.join(1)
        control_port_read_loop.join(1)
        control_device_connection_poller.join(1)

        self.logger.info("Control loops closed")

    def stop(self):
        self.stop_polling()

        if self.device_server is not None:
            self.device_server.close()

        self.session = None
        del self.devices[self.device_id]

        for device in list(self.devices.values()):
            device.disconnect()

        if self.get_value(ATTR_LOOPS_RUNNING):
            self.stop_control_loops()
            self.set_value(ATTR_LOOPS_RUNNING, False)

    @staticmethod
    def start_asyncore_loop():
        control_asyncore_loop.start()

    @staticmethod
    def stop_asyncore_loop():
        control_asyncore_loop.stop()

    @staticmethod
    def start_device_poller_loop():
        control_device_poller.start()

    @staticmethod
    def stop_device_poller_loop():
        control_device_poller.stop()

    @staticmethod
    def start_port_read_loop():
        control_port_read_loop.start()

    @staticmethod
    def stop_port_read_loop():
        control_port_read_loop.stop()

    @staticmethod
    def start_device_connection_poller():
        control_device_connection_poller.start()

    @staticmethod
    def stop_device_connection_poller():
        control_device_connection_poller.stop()

    def set_high_process_priority(self):
        """ Set the priority of the process to below-normal."""
        if psutil.WINDOWS:
            try:
                import win32api, win32process, win32con
            except ImportError:
                self.logger.error("Package pypiwin32 is not installed")
            pid = win32api.GetCurrentProcessId()
            handle = win32api.OpenProcess(win32con.PROCESS_ALL_ACCESS, True, pid)
            win32process.SetPriorityClass(handle, win32process.REALTIME_PRIORITY_CLASS)
            # win32process.SetThreadPriority(win32api.GetCurrentThread(), win32process.THREAD_PRIORITY_TIME_CRITICAL)
        else:
            try:
                process = psutil.Process(os.getpid())
                process.nice(-20)
            except psutil.AccessDenied as e:
                self.logger.error("Could not raise process priority")



from PyQt5.QtCore import pyqtSlot

from kamzik3 import units
from kamzik3.constants import *
from kamzik3.devices.attribute import Attribute
from kamzik3.gui.attributeDeviceDisplayWidget import AttributeDeviceDisplayWidget
from kamzik3.gui.deviceWidget import DeviceWidget
from kamzik3.gui.general.stackableWidget import StackableWidget
from kamzik3.gui.templates.deviceNumAttributeTemplate import Ui_Form


class DeviceNumAttributeWidget(Ui_Form, DeviceWidget, StackableWidget):
    attribute = None
    input_value = None

    def __init__(self, device, attribute, model_image=None, config=None, parent=None):
        self.attribute = Attribute.list_attribute(attribute)
        self.value_controls = []
        DeviceWidget.__init__(self, device, model_image=model_image, config=config, parent=parent)

    def setupUi(self, Form):
        super().setupUi(Form)

        if self.config.get("unit", None) is not None:
            AttributeDeviceDisplayWidget.user_attribute_units[(self.device_id, tuple(self.attribute))] = self.config["unit"]

        self.set_status_label(self.device.get_value(ATTR_STATUS))
        self.label_device_name.setText(self.device_id)

    @pyqtSlot()
    def set_status_label(self, status):
        self.label_status.setText(status)
        self.label_status.setStyleSheet("QLabel {{background:{}}}".format(self.get_status_color(status)))

    @pyqtSlot()
    def slot_handle_configuration(self):
        DeviceWidget.slot_handle_configuration(self)

        self.input_not_ready_placeholder.setParent(None)
        layout = self.widget_holder.layout()
        self.input_value = AttributeDeviceDisplayWidget(self.device_id, self.attribute, config=self.config)
        layout.insertWidget(4, self.input_value)
        self.set_status_label(self.device.get_value(ATTR_STATUS))
        self.value_controls = (
            self.button_minus_double, self.button_minus, self.button_plus, self.button_plus_double,
            self.input_value)

        self.input_value.attribute_widget.sig_unit_changed.connect(self.unit_changed)

    @pyqtSlot("QString", "QString")
    def unit_changed(self, old_units, new_units):
        if old_units is None or new_units is None:
            self.input_step_size.setValue(1)
        else:
            current_value = self.input_step_size.value()
            current_value = units.Quantity(current_value, old_units).to(new_units)
            self.input_step_size.setValue(current_value.m)

    @pyqtSlot("bool")
    def slot_set_enabled(self, value):
        self.setEnabled(value)

    @pyqtSlot("PyQt_PyObject")
    def slot_set_status(self, value):
        DeviceWidget.slot_set_status(self, value)

        if value in READY_DEVICE_STATUSES:
            self.setEnabled(True)
            if value == STATUS_BUSY:
                for widget in self.value_controls:
                    widget.setEnabled(False)
                self.button_stop.setEnabled(True)
            else:
                for widget in self.value_controls:
                    widget.setEnabled(True)
                self.button_stop.setEnabled(False)
        else:
            self.setEnabled(False)

        self.set_status_label(value)

    @pyqtSlot()
    def decrease_value(self):
        step_size = self.input_step_size.value()
        return self.input_value - units.Quantity(step_size, self.input_value.get_unit())

    @pyqtSlot()
    def decrease_value_double(self):
        step_size = self.input_step_size.value()
        return self.input_value - (units.Quantity(step_size, self.input_value.get_unit()) * 10)

    @pyqtSlot()
    def increase_value(self):
        step_size = self.input_step_size.value()
        return self.input_value + units.Quantity(step_size, self.input_value.get_unit())

    @pyqtSlot()
    def increase_value_double(self):
        step_size = self.input_step_size.value()
        return self.input_value + (units.Quantity(step_size, self.input_value.get_unit()) * 10)

    @pyqtSlot()
    def stop(self):
        self.device.stop()

    def close(self):
        self.input_value.close()
        self.input_value = None
        super(DeviceNumAttributeWidget, self).close()



import re
import time

import numpy as np
from bidict import frozenbidict
from pint.errors import UndefinedUnitError

from kamzik3 import DeviceUnitError
from kamzik3.constants import *
from kamzik3.devices.deviceChannel import DeviceChannel
from kamzik3.devices.deviceSocket import DeviceSocket
from kamzik3.snippets.snippetsDecorators import expose_method
from kamzik3.snippets.snippetsUnits import device_units

ATTR_READ_ENCODER = u"Read encoder"

""" Example of yaml configuration
Kohzu0: &Kohzu0 !Device:kamzik3.devices.deviceKohzu.DeviceKohzu
   device_id: Kohzu0
   host: 192.168.83.94
Kohzu0_axis_0: &Kohzu0_axis_0 !Device:kamzik3.devices.deviceKohzu.DeviceKohzuChannel
    device: *Kohzu0
    channel: 0
    device_id: Motor
    config:
     attributes:
       !!python/tuple [Encoder resolution, Value]: 50 nm
       !!python/tuple [Step resolution, Value]: 500 nm
       !!python/tuple [Encoder calc den, Value]: 1
       !!python/tuple [Encoder calc num, Value]: 10
       !!python/tuple [Position, Tolerance]: [0.5um, 0.5um]
       !!python/tuple [Referencing method, Value]: 4
       !!python/tuple [Position +limit, Value]: 10 mm
       !!python/tuple [Position -limit, Value]: -10 mm
       !!python/tuple [Maximum velocity, Value]: 10000
"""


class DeviceKohzu(DeviceSocket):
    terminator = b"\r\n"
    # push_buffer_size = 30
    push_commands_max = 1
    error_description = {
        1: "No STX at the beginning of the command. No header character is required in TCPIP only when RS-232C is used.",
        3: "Characters other than specified characters or number are included.",
        4: "There is no applicable command.",
        5: "The emergency stop signal was detected. It is spontaneously sent from ARIES.",
        6: "Since cutting of Motionnet device was detected while a certain axis was driving, the emergency stop was performed. It is spontaneously sent from ARIES.",
        100: "Total number of parameters is incorrect.",
        101: "Numerical value of the 1st parameter is out of range.",
        102: "Numerical value of the 2nd parameter is out of range.",
        103: "Numerical value of the 3rd parameter is out of range.",
        104: "Numerical value of the 4th parameter is out of range.",
        105: "Numerical value of the 5th parameter is out of range.",
        106: "Numerical value of the 6th parameter is out of range.",
        107: "Numerical value of the 7th parameter is out of range.",
        108: "Numerical value of the 8th parameter is out of range.",
        109: "Numerical value of the 9th parameter is out of range.",
        120: "Number of axes specified in a parameter exceeds controllable number of axes. *Determined according to the connection number of slave controllers and the total number of controllable drive axes.",
        121: "Specified system parameter number is not used.",
        300: "Inside IC related error",
        304: "CW limit is activated during driving and the drive stopped.",
        305: "CCW limit is activated during driving and the drive stopped.",
        306: "One of axes entered limit during multi-axis driving (MPS, SPS) and the drive stopped.",
        307: "Both CW limit and CCW limit are in.",
        308: "The motor is not excited.",
        309: "Tried to operate while axes are driving.",
        310: "Even though the number of oscillation is 1, the coordinate of moving destination specifying the stop time exceeds the range of -134,217,728~134,217,727.",
        311: "Tried to rewrite the pulse counter value of driving axis.",
        312: "Tried to rewrite the encoder counter value of driving axis.",
        313: "Tried to rewrite the system parameter of driving axis.",
        314: "Because emergency stop is detected, driving axes are stopped.",
        315: "Because alarm is detected, driving axes are stopped.",
        316: "- side soft limit is more than + side soft limit.",
        317: "Due to + side soft limit, the drive is stopped.",
        318: "Due to - side soft limit, the drive is stopped.",
        319: "One of axes entered soft limit during multi-axis driving (MPS, SPS) and the drive is stopped.",
        320: "Movement amount of standalone axis is 0 between the linear interpolations.",
        321: "Tried to make it operate in the state where a servoready signal is not ON.",
        322: "Since cutting of Motionnet device was detected while the axis was driving, the axis was stopped. Feedback 401 Though the number of retry_count counts exceeded in feedback control, theencoder feedback did not complete.",
        500: "MPI command is not issued.",
        501: "Drive parameter corresponding to the 1st axis of MPS command is not set.",
        502: "Drive parameter corresponding to the 2nd axis of MPS command is not set.",
        503: "Drive parameter corresponding to the 3rd axis of MPS command is not set.",
        504: "Drive parameter corresponding to the 4th axis of MPS command is not set.",
        505: "The coordinate at the movement destination of the MPS 1st axis is out of range (-13.,'21.,'728 to 13.,'21.,'727).",
        506: "The coordinate at the movement destination of the MPS 2nd axis is out of range (-13.,'21.,'728 to 13.,'21.,'727).",
        507: "The coordinate at the movement destination of the MPS 3rd axis is out of range (-13.,'21.,'728 to 13.,'21.,'727).",
        508: "The coordinate at the movement destination of the MPS 4th axis is out of range (-13.,'21.,'728 to 13.,'21.,'727).",
        510: "3 or more axes specified for simultaneous drive are the same.",
        511: "1st and 2nd axes specified for simultaneous drive are the same.",
        512: "1st and 3rd axes specified for simultaneous drive are the same.",
        513: "1st and 4th axes specified for simultaneous drive are the same.",
        514: "2nd and 3rd axes specified for simultaneous drive are the same.",
        515: "2nd and 4th axes specified for simultaneous drive are the same.",
        516: "3rd and 4th axes specified for simultaneous drive are the same.",
        600: "Failed to create a speed table.",
        601: "Acceleration time is large.",
        602: "Acceleration time is small.",
        603: "Deceleration time is large.",
        604: "Deceleration time is small.",
        605: "Start speed exceeds 50% of the maximum speed.",
        606: "Maximum speed of slave axis between the linear interpolations exceeds the limit value (SYS No.16).",
        607: "Maximum speed exceeds the limit value (SYS No.16).",
        700: "While outputting the trigger signal, It was going to change the trigger-related system parameter(SYS.No.51 to 56).",
        701: "TRS command is issued for driving axes.",
        702: "Trigger output doesn't stop after exceeding the setting time.",
        703: "Trigger output stopped before exceeding the setting time.",
        800: "Because emergency stop is detected a command is not executed. It restores by issue of the REM command.",
        801: "Because emergency stop is detected a release cannot be executed. Please publish the REM command after removing the factor of an emergency stop.",
        802: "A command is not executed for the emergency stop state by connection interception (power supply cutting is included) of a Motionnet device. It restores by issue of the RAX command.",
        900: "RAX command is executed during driving.",
        901: "While the axis drove, the WIP command or the PIP command was published.",
        50: "Change of Motionnet device composition was detected. It is spontaneously sent from ARIE.",
        350: "Moving target position exceeds the soft limit. This warning is returned when the drive command's response method is quick while the soft limit is valid and the target position exceeds the soft limit. Driving continues to the soft limit."
    }
    feedback_control_method = frozenbidict(
        {u"No correction": 0, u"Correction at positioning": 1, u"Normal correction": 2})
    trigger_source = frozenbidict(
        {u"Motor pulse": 0, u"Encoder pulse x1": 1, u"Encoder pulse x2": 2, u"Encoder pulse x3": 3})
    trigger_pulse_width = frozenbidict({u"1 usec": 1, u"10 usec": 2, u"100 usc": 3, u"1000 usc": 4})
    trigger_edge = frozenbidict({RISING: 0, FALLING: 1})
    trigger_logic = frozenbidict({POSITIVE: 0, NEGATIVE: 1})
    microstep_select = frozenbidict({u"M1": 0, u"M2": 1})
    motor_selection = frozenbidict({u"Pulse": 0, u"Servo": 1})
    stop_type = frozenbidict({u"Deceleration and stop": 0, u"Emergency stop": 1})
    motor_excitation = frozenbidict({u"Off": 0, u"On": 1})
    accelerating_pattern = frozenbidict({u"Rectangular drive": 1, u"Trapezoidal drive": 2, u"S-shaped drive": 3})
    encoder_adding = frozenbidict({u"Normal": 0, u"Reverse": 1})
    system_setting = frozenbidict({ATTR_REFERENCING_METHOD: 2,
                                   ATTR_ROTATION_CHANGE: 7,
                                   ATTR_SWAP_LIMITS: 8,
                                   ATTR_POSITION_UPPER_LIMIT: 14,
                                   ATTR_POSITION_LOWER_LIMIT: 15,
                                   ATTR_MAXIMUM_VELOCITY: 16,
                                   ATTR_ENCODER_CALC_NUM: 33,
                                   ATTR_ENCODER_CALC_DEN: 34,
                                   ATTR_ENCODER_ADDING: 35,
                                   ATTR_FEEDBACK_CONTROL_TYPE: 41,
                                   ATTR_RETRY_COUNT: 43,
                                   ATTR_FEEDBACK_TIMEOUT: 44,
                                   ATTR_TRIGGER_SOURCE: 51,
                                   ATTR_TRIGGER_EDGE: 52,
                                   ATTR_TRIGGER_PM_PITCH: 53,
                                   ATTR_TRIGGER_ENC_PITCH: 54,
                                   ATTR_TRIGGER_PULSE_WIDTH: 55,
                                   ATTR_TRIGGER_LOGIC: 56,
                                   ATTR_MOTOR_EXCITATION: 61,
                                   ATTR_MOTOR_SELECTION: 62,
                                   ATTR_MICROSTEP_SELECT: 65,
                                   ATTR_STOP_TYPE: 99
                                   })

    def __init__(self, host, port=12321, device_id=None, config=None):
        super(DeviceKohzu, self).__init__(host, port, device_id, config)

    def _init_attributes(self):
        super(DeviceKohzu, self)._init_attributes()

        self.create_attribute(ATTR_SERIAL_NUMBER, readonly=True)
        self.create_attribute(ATTR_CHANNELS, readonly=True, default_type=np.uint16)
        self.create_attribute(ATTR_MODULES, readonly=True, default_type=np.uint16)

    def valid_command_format(self, command):
        if command is not None and command[-1] == "\n":
            return True
        else:
            return False

    def handle_configuration(self):
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        # Get ARIES serial number
        self.command(u"IDN\r\n")
        # Get ARIES configuration
        self.command(u"RAX\r\n", callback=_finish_configuration)

    def handle_readout(self, readout_buffer):
        command, output, callback, token = super(DeviceKohzu, self).handle_readout(readout_buffer)
        output_parts = output.split("\t")
        response_flag = output_parts.pop(0)

        if response_flag == "E":
            command, error_code = output_parts
            readout_buffer = self.error_description.get(int(error_code), u"Undefined error")
            self.handle_command_error(command, readout_buffer)
            s = re.search("[A-Z]+([0-9]+)", command)
            command = RESPONSE_ERROR
            if s is not None and s != -1:
                channel = int(s.group(1)) - 1
                self.notify((channel, ATTR_LAST_ERROR), readout_buffer)
        elif command == "IDN\r\n":
            self.set_value(ATTR_SERIAL_NUMBER, "{} {}".format(" ".join(output_parts[:2]), ".".join(output_parts[2:])))
        elif command == "RAX\r\n":
            self.set_value(ATTR_MODULES, int(output_parts[1]))
            self.set_value(ATTR_CHANNELS, int(output_parts[2]))
        elif output_parts[0][:3] == "RSY":
            channel = int(output_parts[0][3:]) - 1
            self.notify((channel, self.system_setting.inv[int(output_parts[1])]), int(output_parts[2]))
        elif output_parts[0][:3] == "RTB":
            channel = int(output_parts[0][3:]) - 1
            self.notify((channel, ATTR_START_SPEED), int(output_parts[2]))
            self.notify((channel, ATTR_TOP_SPEED), int(output_parts[3]))
            self.notify((channel, ATTR_ACCELERATION_TIME), int(output_parts[4]))
            self.notify((channel, ATTR_DECELERATION_TIME), int(output_parts[5]))
            self.notify((channel, ATTR_ACCELERATION_PATTERN), self.accelerating_pattern.inv[int(output_parts[6])])
        elif output_parts[0][:3] == "STR":
            channel = int(output_parts[0][3:]) - 1
            self.notify((channel, ATTR_STATE), output_parts[1:])
        elif output_parts[0][:3] == "RDE":
            channel = int(output_parts[0][3:]) - 1
            self.notify((channel, ATTR_ENCODER_READ), int(output_parts[1]))
        elif output_parts[0][:3] == "RDP":
            channel = int(output_parts[0][3:]) - 1
            self.notify((channel, ATTR_PULSE_READ), int(output_parts[1]))
        elif output_parts[0][:3] == "ROG":
            channel = int(output_parts[0][3:]) - 1
            self.notify((channel, ATTR_REFERENCED), int(output_parts[1]))

        if callback is not None:
            self.handle_readout_callback(callback, command, readout_buffer)
        if token:
            self.device_server.push_message(self.device_id, [command, readout_buffer], token)
            for observer in self._observers[:]:
                if isinstance(observer, DeviceKohzuChannel):
                    self.device_server.push_message(observer.device_id, [command, readout_buffer], token)


class DeviceKohzuChannel(DeviceChannel):

    def _init_attributes(self):
        super(DeviceKohzuChannel, self)._init_attributes()

        self.create_attribute(ATTR_STATE, readonly=True, default_type=np.uint8, min_value=0, max_value=3)
        self.create_attribute(ATTR_EMERGENCY_STOP, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_CWL, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_CCWL, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_ORG, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_NORG, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_REFERENCED, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_ROTATION_CHANGE, default_type=np.uint8, min_value=0, max_value=1,
                              set_function=lambda v: self.set_sys(ATTR_ROTATION_CHANGE, v))
        self.create_attribute(ATTR_SWAP_LIMITS, default_type=np.uint8, min_value=0, max_value=1,
                              set_function=lambda v: self.set_sys(ATTR_SWAP_LIMITS, v))
        self.create_attribute(ATTR_ENCODER_READ, default_type=np.int64, set_function=self.set_encoder_value)
        self.create_attribute(ATTR_PULSE_READ, default_type=np.int64, set_function=self.set_pulse_value, unit="pulse")
        self.create_attribute(ATTR_POSITION, default_type=np.int64, set_function=self._move_absolute, unit="nm",
                              set_value_when_set_function=False)
        self.create_attribute(ATTR_REFERENCING_METHOD, default_type=np.uint8, min_value=1, max_value=15,
                              set_function=lambda v: self.set_sys(ATTR_REFERENCING_METHOD, v))
        self.create_attribute(ATTR_POSITION_LOWER_LIMIT, default_type=np.int64, min_value=-134217728,
                              max_value=134217727, unit="nm",
                              set_function=lambda v: self.set_sys(ATTR_POSITION_LOWER_LIMIT, v))
        self.create_attribute(ATTR_STEP_RESOLUTION, default_value=500, default_type=np.int64, min_value=1,
                              max_value=134217727, unit="nm")
        self.create_attribute(ATTR_START_SPEED, default_value=1, default_type=np.int32, min_value=1, max_value=2.5e6,
                              unit="pulse/s",
                              set_function=lambda v: self.write_speed_table(start_speed=v))
        self.create_attribute(ATTR_TOP_SPEED, default_value=1, default_type=np.int32, min_value=2, max_value=5e6,
                              unit="pulse/s",
                              set_function=lambda v: self.write_speed_table(top_speed=v))
        self.create_attribute(ATTR_ACCELERATION_TIME, default_value=1, default_type=np.int32, min_value=1,
                              max_value=10e3, unit="ms",
                              set_function=lambda v: self.write_speed_table(acceleration_time=v))
        self.create_attribute(ATTR_DECELERATION_TIME, default_value=1, default_type=np.int32, min_value=1,
                              max_value=10e3, unit="ms",
                              set_function=lambda v: self.write_speed_table(deceleration_time=v))
        self.create_attribute(ATTR_ENCODER_RESOLUTION, default_value=50, default_type=np.int64, min_value=1,
                              max_value=134217727, unit="nm")
        self.create_attribute(ATTR_POSITION_UPPER_LIMIT, default_type=np.int64, min_value=-134217728,
                              max_value=134217727, unit="nm",
                              set_function=lambda v: self.set_sys(ATTR_POSITION_UPPER_LIMIT, v))
        self.create_attribute(ATTR_MAXIMUM_VELOCITY, default_type=np.uint32, min_value=2, max_value=5000000,
                              set_function=lambda v: self.set_sys(ATTR_MAXIMUM_VELOCITY, v))
        self.create_attribute(ATTR_ENCODER_CALC_NUM, default_type=np.uint32, min_value=1, max_value=134217727,
                              set_function=lambda v: self.set_sys(ATTR_ENCODER_CALC_NUM, v))
        self.create_attribute(ATTR_ENCODER_CALC_DEN, default_type=np.uint32, min_value=1, max_value=134217727,
                              set_function=lambda v: self.set_sys(ATTR_ENCODER_CALC_DEN, v))
        self.create_attribute(ATTR_FEEDBACK_CONTROL_TYPE, default_type=DeviceKohzu.feedback_control_method.keys(),
                              set_function=lambda v: self.set_sys(ATTR_FEEDBACK_CONTROL_TYPE,
                                                                  DeviceKohzu.feedback_control_method[v]))
        self.create_attribute(ATTR_RETRY_COUNT, default_type=np.uint16, min_value=1, max_value=10e3,
                              set_function=lambda v: self.set_sys(ATTR_RETRY_COUNT, v))
        self.create_attribute(ATTR_FEEDBACK_TIMEOUT, default_type=np.uint16, min_value=1, max_value=10e3, unit="ms",
                              set_function=lambda v: self.set_sys(ATTR_FEEDBACK_TIMEOUT, v))
        self.create_attribute(ATTR_TRIGGER_SOURCE, default_type=DeviceKohzu.trigger_source.keys(),
                              set_function=lambda v: self.set_sys(ATTR_TRIGGER_SOURCE, DeviceKohzu.trigger_source[v]))
        self.create_attribute(ATTR_TRIGGER_EDGE, default_type=DeviceKohzu.trigger_edge.keys(),
                              set_function=lambda v: self.set_sys(ATTR_TRIGGER_EDGE, DeviceKohzu.trigger_edge[v]))
        self.create_attribute(ATTR_TRIGGER_PM_PITCH, default_type=np.uint32, min_value=1, max_value=100e3,
                              set_function=lambda v: self.set_sys(ATTR_TRIGGER_PM_PITCH, v))
        self.create_attribute(ATTR_TRIGGER_ENC_PITCH, default_type=np.uint32, min_value=1, max_value=100e3,
                              set_function=lambda v: self.set_sys(ATTR_TRIGGER_ENC_PITCH, v))
        self.create_attribute(ATTR_TRIGGER_PULSE_WIDTH, default_type=DeviceKohzu.trigger_pulse_width.keys(),
                              set_function=lambda v: self.set_sys(ATTR_TRIGGER_PULSE_WIDTH,
                                                                  DeviceKohzu.trigger_pulse_width[v]))
        self.create_attribute(ATTR_ENCODER_ADDING, default_type=DeviceKohzu.encoder_adding.keys(),
                              set_function=lambda v: self.set_sys(ATTR_ENCODER_ADDING, DeviceKohzu.encoder_adding[v]))
        self.create_attribute(ATTR_TRIGGER_LOGIC, default_type=DeviceKohzu.trigger_logic.keys(),
                              set_function=lambda v: self.set_sys(ATTR_TRIGGER_LOGIC, DeviceKohzu.trigger_logic[v]))
        self.create_attribute(ATTR_MOTOR_EXCITATION, default_type=DeviceKohzu.motor_excitation.keys(),
                              set_function=lambda v: self.set_sys(ATTR_MOTOR_EXCITATION,
                                                                  DeviceKohzu.motor_excitation[v]))
        self.create_attribute(ATTR_MOTOR_SELECTION, default_type=DeviceKohzu.motor_selection.keys(),
                              set_function=lambda v: self.set_sys(ATTR_MOTOR_SELECTION, DeviceKohzu.motor_selection[v]))
        self.create_attribute(ATTR_MICROSTEP_SELECT, default_type=DeviceKohzu.microstep_select.keys(),
                              set_function=lambda v: self.set_sys(ATTR_MICROSTEP_SELECT,
                                                                  DeviceKohzu.microstep_select[v]))
        self.create_attribute(ATTR_STOP_TYPE, default_type=DeviceKohzu.stop_type.keys(),
                              set_function=lambda v: self.set_sys(ATTR_STOP_TYPE, DeviceKohzu.stop_type[v]))
        self.create_attribute(ATTR_ACCELERATION_PATTERN, default_type=DeviceKohzu.accelerating_pattern.keys(),
                              set_function=lambda v: self.write_speed_table(acceleration_pattern=v))
        self.create_attribute(ATTR_READ_ENCODER, default_value=True, readonly=False, default_type=np.bool)
        self.create_attribute(ATTR_VELOCITY, default_value=0, default_type=np.int64, min_value=0, max_value=3e6,
                              unit="nm/s", set_function=self.set_velocity)

    def handle_configuration(self):
        if self.configured:
            return

        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.configured = True
            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        self.connected = True
        self.get_system_setting(ATTR_POSITION_LOWER_LIMIT)
        self.get_system_setting(ATTR_POSITION_UPPER_LIMIT)
        self.get_system_setting(ATTR_ROTATION_CHANGE)
        self.get_system_setting(ATTR_SWAP_LIMITS)
        self.get_system_setting(ATTR_MAXIMUM_VELOCITY)
        self.get_system_setting(ATTR_ENCODER_CALC_NUM)
        self.get_system_setting(ATTR_ENCODER_CALC_DEN)
        self.get_system_setting(ATTR_FEEDBACK_CONTROL_TYPE)
        self.get_system_setting(ATTR_RETRY_COUNT)
        self.get_system_setting(ATTR_TRIGGER_SOURCE)
        self.get_system_setting(ATTR_TRIGGER_EDGE)
        self.get_system_setting(ATTR_TRIGGER_PM_PITCH)
        self.get_system_setting(ATTR_TRIGGER_ENC_PITCH)
        self.get_system_setting(ATTR_TRIGGER_PULSE_WIDTH)
        self.get_system_setting(ATTR_TRIGGER_LOGIC)
        self.get_system_setting(ATTR_MOTOR_EXCITATION)
        self.get_system_setting(ATTR_MOTOR_SELECTION)
        self.get_system_setting(ATTR_MICROSTEP_SELECT)
        self.get_system_setting(ATTR_STOP_TYPE)
        self.get_system_setting(ATTR_REFERENCING_METHOD)
        self.get_system_setting(ATTR_ENCODER_ADDING)
        self.get_referenced()
        self.command(u"RTB{}/0\r\n".format(self.channel + 1), callback=_finish_configuration)

    def set_velocity(self, value):
        value = int(value / self[ATTR_STEP_RESOLUTION][VALUE])
        self.set_value(ATTR_START_SPEED, 2)
        self.set_value(ATTR_TOP_SPEED, 10)
        self.set_value(ATTR_TOP_SPEED, value)
        self.set_value(ATTR_START_SPEED, int(value / 2.2))

    def subject_update(self, key, value, subject):
        super(DeviceKohzuChannel, self).subject_update(key, value, subject)

        if self.connected and isinstance(key, tuple) and key[0] == self.channel:
            attribute = key[1]
            if attribute == ATTR_FEEDBACK_CONTROL_TYPE:
                value = DeviceKohzu.feedback_control_method.inv[value]
            elif attribute == ATTR_TRIGGER_EDGE:
                value = DeviceKohzu.trigger_edge.inv[value]
            elif attribute == ATTR_TRIGGER_SOURCE:
                value = DeviceKohzu.trigger_source.inv[value]
            elif attribute == ATTR_TRIGGER_LOGIC:
                value = DeviceKohzu.trigger_logic.inv[value]
            elif attribute == ATTR_TRIGGER_PULSE_WIDTH:
                value = DeviceKohzu.trigger_pulse_width.inv[value]
            elif attribute == ATTR_ENCODER_ADDING:
                value = DeviceKohzu.encoder_adding.inv[value]
            elif attribute == ATTR_MOTOR_EXCITATION:
                value = DeviceKohzu.motor_excitation.inv[value]
            elif attribute == ATTR_MOTOR_SELECTION:
                value = DeviceKohzu.motor_selection.inv[value]
            elif attribute == ATTR_MICROSTEP_SELECT:
                value = DeviceKohzu.microstep_select.inv[value]
            elif attribute == ATTR_STOP_TYPE:
                value = DeviceKohzu.stop_type.inv[value]
            elif attribute == ATTR_ENCODER_READ and self.get_value(ATTR_READ_ENCODER):
                self.set_raw_value(ATTR_POSITION, value * self[ATTR_ENCODER_RESOLUTION][VALUE])
            elif attribute == ATTR_PULSE_READ and not self.get_value(ATTR_READ_ENCODER):
                self.set_raw_value(ATTR_POSITION, value * self[ATTR_STEP_RESOLUTION][VALUE])
            elif attribute == ATTR_POSITION_LOWER_LIMIT:
                self.set_attribute([ATTR_POSITION, MIN], value)
            elif attribute == ATTR_POSITION_UPPER_LIMIT:
                self.set_attribute([ATTR_POSITION, MAX], value)
            elif attribute == ATTR_STATE:
                if int(value[0]) == 0:
                    self.set_status(STATUS_IDLE)
                else:
                    self.set_status(STATUS_BUSY)
                self.set_value(ATTR_EMERGENCY_STOP, bool(int(value[1])))

                org_value = int(value[2])
                if org_value == 0:
                    self.set_value(ATTR_ORG, False)
                    self.set_value(ATTR_NORG, False)
                elif org_value == 1:
                    self.set_value(ATTR_ORG, False)
                    self.set_value(ATTR_NORG, True)
                elif org_value == 2:
                    self.set_value(ATTR_ORG, True)
                    self.set_value(ATTR_NORG, False)
                elif org_value == 3:
                    self.set_value(ATTR_ORG, True)
                    self.set_value(ATTR_NORG, True)

                limit_value = int(value[3])
                if limit_value == 0:
                    self.set_value(ATTR_CWL, False)
                    self.set_value(ATTR_CCWL, False)
                elif limit_value == 1:
                    self.set_value(ATTR_CWL, False)
                    self.set_value(ATTR_CCWL, True)
                elif limit_value == 2:
                    self.set_value(ATTR_CWL, True)
                    self.set_value(ATTR_CCWL, False)
                elif limit_value == 3:
                    self.set_value(ATTR_CWL, True)
                    self.set_value(ATTR_CCWL, True)

                value = int(value[0])

            self.set_raw_value(ATTR_LATENCY, self.device.get_value(ATTR_LATENCY))
            self.set_raw_value(attribute, value)

    def start_polling(self):
        super(DeviceKohzuChannel, self).start_polling()
        self.poll_command("STR{}\r\n".format(self.channel + 1), 300)
        self.poll_command("RDE{}\r\n".format(self.channel + 1), 300)
        self.poll_command("RDP{}\r\n".format(self.channel + 1), 300)
        self.poll_command("ROG{}\r\n".format(self.channel + 1), 500)

    def stop_polling(self):
        self.remove_poll_command("STR{}\r\n".format(self.channel + 1), 300)
        self.remove_poll_command("RDE{}\r\n".format(self.channel + 1), 300)
        self.remove_poll_command("RDP{}\r\n".format(self.channel + 1), 300)
        self.remove_poll_command("ROG{}\r\n".format(self.channel + 1), 500)

    def get_system_setting(self, setting, callback=None):
        return self.command(u"RSY{}/{}\r\n".format(self.channel + 1, DeviceKohzu.system_setting[setting]), callback,
                            with_token=True)

    def set_sys(self, setting, value, callback=None):
        # if setting == ATTR_POSITION_UPPER_LIMIT:
        #     print(value)
        token = self.command(
            "WSY{}/{}/{}\r\n".format(self.channel + 1, DeviceKohzu.system_setting[setting], int(value)), callback,
            with_token=True)
        self.get_system_setting(setting)
        if setting == ATTR_POSITION_UPPER_LIMIT:
            self.get_system_setting(ATTR_POSITION_LOWER_LIMIT)
        elif setting == ATTR_POSITION_LOWER_LIMIT:
            self.get_system_setting(ATTR_POSITION_UPPER_LIMIT)
        return token

    def write_speed_table(self, start_speed=None, top_speed=None, acceleration_time=None, deceleration_time=None,
                          acceleration_pattern=None, callback=None):
        if start_speed is None:
            start_speed = self.get_value(ATTR_START_SPEED)
        if top_speed is None:
            top_speed = self.get_value(ATTR_TOP_SPEED)
        if acceleration_time is None:
            acceleration_time = self.get_value(ATTR_ACCELERATION_TIME)
        if deceleration_time is None:
            deceleration_time = self.get_value(ATTR_DECELERATION_TIME)
        if acceleration_pattern is None:
            acceleration_pattern = self.get_value(ATTR_ACCELERATION_PATTERN)

        top_speed = int(top_speed)
        start_speed = int(start_speed)
        acceleration_time = int(acceleration_time)
        deceleration_time = int(deceleration_time)
        acceleration_pattern = DeviceKohzu.accelerating_pattern[acceleration_pattern]

        self.command(
            "WTB{}/{}/{}/{}/{}/{}/{}\r\n".format(self.channel + 1, 0, start_speed, top_speed, acceleration_time,
                                                 deceleration_time, acceleration_pattern), callback)
        return self.command(u"RTB{}/0\r\n".format(self.channel + 1), callback, with_token=True)

    @expose_method()
    def stop(self, callback=None):
        self.logger.info(u"Stop movement")
        stop_type = DeviceKohzu.stop_type[self.get_value(ATTR_STOP_TYPE)]
        return self.command(u"STP{}/{}\r\n".format(self.channel + 1, stop_type), callback, with_token=True)

    @expose_method()
    def system_rest(self, callback=None):
        self.logger.info(u"System rest")
        return self.command(u"RST\r\n", callback, with_token=True)

    @expose_method()
    def find_reference(self, callback=None):
        self.logger.info(u"Reference")
        return self.command(u"ORG{}/{}/{}\r\n".format(self.channel + 1, 0, 1), callback, with_token=True)

    @expose_method()
    def find_reference_reversed(self, callback=None):
        self.stop_polling()
        rotational_direction = not self[ATTR_ROTATION_CHANGE][VALUE]
        swap_limits = not self[ATTR_SWAP_LIMITS][VALUE]

        self.set_value(ATTR_ROTATION_CHANGE, rotational_direction)
        self.set_value(ATTR_SWAP_LIMITS, swap_limits)

        def _swap_rotation_back(*_, **__):
            self.start_polling()
            self.set_value(ATTR_ROTATION_CHANGE, int(not rotational_direction))
            self.set_value(ATTR_SWAP_LIMITS, int(not swap_limits))
            if callback is not None:
                callback()

        self.logger.info(u"Reference")
        return self.command(u"ORG{}/{}/{}\r\n".format(self.channel + 1, 0, 0), _swap_rotation_back, with_token=True)

    @expose_method()
    def release_emergency_stop(self, callback=None):
        self.logger.info(u"Emergency stop released")
        return self.command(u"REM\r\n", callback, with_token=True)

    @expose_method()
    def jog_clockwise(self, callback=None):
        self.logger.info(u"Jog clockwise")
        return self.command(u"FRP{}/{}/{}\r\n".format(self.channel + 1, 0, 0), callback, with_token=True)

    @expose_method()
    def jog_counter_clockwise(self, callback=None):
        self.logger.info(u"Jog clockwise")
        return self.command(u"FRP{}/{}/{}\r\n".format(self.channel + 1, 0, 1), callback, with_token=True)

    @expose_method({"position": ATTR_POSITION})
    def move_absolute(self, position, callback=None):
        try:
            position = device_units(self, ATTR_POSITION, position)
            self.logger.info(u"Move to absolute position {}".format(position))
            return self.set_value(ATTR_POSITION, position, callback=callback)
        except (UndefinedUnitError, ValueError) as e:
            self.logger.exception(e)
            raise DeviceUnitError(e)

    @expose_method({"position": ATTR_POSITION})
    def move_relative(self, position, callback=None):
        try:
            position = device_units(self, ATTR_POSITION, position)
            self.logger.info(u"Move by relative position {}".format(position))
            return self.set_value(ATTR_POSITION, self[ATTR_POSITION].value() + position, callback=callback)
        except (UndefinedUnitError, ValueError) as e:
            self.logger.exception(e)
            raise DeviceUnitError(e)

    def _move_absolute(self, value, callback=None):
        self.logger.info(u"Move to position {}".format(value))
        # APS{channel},{speed table #},{position},{response type}
        value /= self[ATTR_STEP_RESOLUTION][VALUE]
        return self.command(u"APS{}/{}/{}/{}\r\n".format(self.channel + 1, 0, int(value), 1), callback, with_token=True)

    def _move_relative(self, value, callback=None):
        self.logger.info(u"Move to position {}".format(value))
        # APS{channel},{speed table #},{position},{response type}
        value /= self[ATTR_STEP_RESOLUTION][VALUE]
        return self.command(u"RPS{}/{}/{}/{}\r\n".format(self.channel + 1, 0, int(value), 1), callback, with_token=True)

    def get_referenced(self, callback=None):
        return self.command(u"ROG{}\r\n".format(self.channel + 1), callback, with_token=True)

    def set_encoder_value(self, value, callback=None):
        value = int(value)
        self.logger.info(u"Set encoder value to {}".format(value))
        return self.command(u"WRE{}/{}\r\n".format(self.channel + 1, value), callback, with_token=True)

    def set_pulse_value(self, value, callback=None):
        value = int(value)
        self.logger.info(u"Set pulse value to {}".format(value))
        return self.command(u"WRP{}/{}\r\n".format(self.channel + 1, value), callback, with_token=True)



from PyQt5 import QtWidgets
from PyQt5.QtCore import pyqtSlot, Qt
from PyQt5.QtWidgets import QLabel, QHBoxLayout, QWidget, QPushButton, QSizePolicy, QLineEdit

from kamzik3 import DeviceError, units
from kamzik3.constants import *
from kamzik3.gui.attributeDisplayWidget import AttributeDisplayWidget
from kamzik3.gui.deviceWidget import DeviceWidget
from kamzik3.gui.general.stackableWidget import StackableWidget
from kamzik3.gui.templates.deviceMethodTemplate import Ui_Form
from kamzik3.snippets.snippetsWidgets import show_error_message


class DeviceMethodWidget(Ui_Form, DeviceWidget, StackableWidget):
    attribute = None
    input_value = None

    def __init__(self, device, method, model_image=None, config=None, parent=None):
        self.method = method
        self.value_controls = ()
        DeviceWidget.__init__(self, device, model_image=model_image, config=config, parent=parent)
        self.set_status_label(self.device.get_value(ATTR_STATUS))
        self.label_device_name.setText(self.device_id)

    @pyqtSlot()
    def set_status_label(self, status):
        self.label_status.setText(status)
        self.label_status.setStyleSheet("QLabel {{background:{}}}".format(self.get_status_color(status)))

    @pyqtSlot()
    def slot_handle_configuration(self):
        DeviceWidget.slot_handle_configuration(self)

        self.input_not_ready_placeholder.setParent(None)
        layout = self.widget_holder.layout()
        self.input_value = self.method_widget()
        layout.insertWidget(2, self.input_value)
        self.set_status_label(self.device.get_value(ATTR_STATUS))
        self.value_controls = (self.input_value,)

    def method_widget(self):
        device_methods = self.device.exposed_methods
        method_widget = QWidget()

        def exec_method(method_name, argument_inputs=None):
            kwargs = {}
            for argument_input in argument_inputs:
                if isinstance(argument_input, AttributeDisplayWidget):
                    value = argument_input.get_attribute_value()
                    if isinstance(value, units.Quantity):
                        kwargs[argument_input.name] = "{:~}".format(value)
                    else:
                        kwargs[argument_input.name] = value
            try:
                getattr(self.device, method_name)(**kwargs)
            except DeviceError as e:
                show_error_message(e, parent=self)

        layout = QHBoxLayout(method_widget)
        layout.setContentsMargins(0, 0, 0, 0)
        for row, (method_name, method_attributes) in enumerate(device_methods):
            if method_name == self.method:
                method_label = QLabel(method_name + " (")
                method_label.setAlignment(Qt.AlignCenter | Qt.AlignRight)
                method_label.setStyleSheet("QLabel {font-weight:bold}")
                layout.addWidget(method_label)
                method_execute_button = QPushButton("< Call >")
                method_execute_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
                argument_inputs = []
                if method_attributes:
                    inputs_widget = QWidget()
                    inputs_layout = QHBoxLayout()
                    inputs_widget.setLayout(inputs_layout)
                    for attribute_title, attribute_type in method_attributes.items():
                        device_attribute = self.device.get_attribute(attribute_type)
                        if device_attribute is None:
                            inputs_layout.addWidget(QLabel("{}".format(attribute_title)))
                            input_value = QLineEdit()
                            input_value.setPlaceholderText(attribute_type)
                            input_value.name = attribute_title
                            inputs_layout.addWidget(input_value)
                            argument_inputs.append(input_value)
                        else:
                            attribute_widget = AttributeDisplayWidget(attribute_title, device_attribute)
                            attribute_widget.name = attribute_title
                            inputs_layout.addWidget(attribute_widget.label_widget)
                            inputs_layout.addWidget(attribute_widget.input_widget)
                            inputs_layout.addWidget(attribute_widget.unit_widget)
                            argument_inputs.append(attribute_widget)
                    layout.addWidget(inputs_widget)
                else:
                    label = QLabel("None")
                    label.setAlignment(Qt.AlignCenter | Qt.AlignHCenter)
                    label.setStyleSheet("QLabel {font-weight:bold}")
                    layout.addWidget(label)

                method_execute_button.released.connect(
                    lambda method_name=method_name, argument_inputs=argument_inputs: exec_method(method_name,
                                                                                                 argument_inputs))
                method_label = QLabel(")")
                method_label.setStyleSheet("QLabel {font-weight:bold}")
                layout.addWidget(method_label)
                layout.addWidget(method_execute_button)
                break

        return method_widget

    @pyqtSlot("bool")
    def slot_set_enabled(self, value):
        self.setEnabled(value)

    @pyqtSlot("PyQt_PyObject")
    def slot_set_status(self, value):
        DeviceWidget.slot_set_status(self, value)

        if value in READY_DEVICE_STATUSES:
            self.setEnabled(True)
            if value == STATUS_BUSY:
                for widget in self.value_controls:
                    widget.setEnabled(False)
                self.button_stop.setEnabled(True)
            else:
                for widget in self.value_controls:
                    widget.setEnabled(True)
                self.button_stop.setEnabled(False)
        else:
            self.setEnabled(False)

        self.set_status_label(value)

    @pyqtSlot()
    def stop(self):
        self.device.stop()

    def close(self):
        self.input_value.close()
        self.input_value = None
        super(DeviceMethodWidget, self).close()

class DeviceSimpleMethodsListWidget(DeviceMethodWidget):

    def __init__(self, device, methods_list, model_image=None, config=None, parent=None):
        self.methods_list = methods_list
        self.value_controls = []
        self.input_values = []
        DeviceWidget.__init__(self, device, model_image=model_image, config=config, parent=parent)
        self.set_status_label(self.device.get_value(ATTR_STATUS))
        self.label_device_name.setText(self.device_id)

    @pyqtSlot()
    def slot_handle_configuration(self):
        DeviceWidget.slot_handle_configuration(self)

        self.input_not_ready_placeholder.setParent(None)
        layout = self.widget_holder.layout()
        layout.insertStretch(2)
        for required_method in reversed(self.methods_list):
            input_value = self.method_widget(required_method)
            layout.insertWidget(2, input_value)
            self.set_status_label(self.device.get_value(ATTR_STATUS))
            self.value_controls.append(input_value)
            self.input_values.append(input_value)


    def method_widget(self, required_method_name):
        device_methods = self.device.exposed_methods
        method_widget = QWidget()

        def exec_method(method_name, argument_inputs=None):
            kwargs = {}
            for argument_input in argument_inputs:
                if isinstance(argument_input, AttributeDisplayWidget):
                    value = argument_input.get_attribute_value()
                    if isinstance(value, units.Quantity):
                        kwargs[argument_input.name] = "{:~}".format(value)
                    else:
                        kwargs[argument_input.name] = value
            try:
                getattr(self.device, method_name)(**kwargs)
            except DeviceError as e:
                show_error_message(e, parent=self)

        layout = QHBoxLayout(method_widget)
        layout.setContentsMargins(0, 0, 0, 0)
        for row, (method_name, method_attributes) in enumerate(device_methods):
            if method_name == required_method_name:
                method_execute_button = QPushButton(method_name)
                method_execute_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
                argument_inputs = []

                method_execute_button.released.connect(
                    lambda method_name=method_name, argument_inputs=argument_inputs: exec_method(method_name,
                                                                                                 argument_inputs))
                layout.addWidget(method_execute_button)
                break

        return method_widget

    def close(self):
        for input in self.input_values:
            input.close()
        self.input_values = None
        super(DeviceMethodWidget, self).close()


import logging
import time
from collections import deque
from copy import copy
from math import inf
from threading import Thread

import numpy as np
import serial
from pint import UndefinedUnitError, DimensionalityError

import kamzik3
from kamzik3 import WriteException, CommandFormatException, DeviceError
from kamzik3.constants import *
from kamzik3.devices.attribute import Attribute, SetFunction
from kamzik3.devices.subject import Subject
from kamzik3.snippets.snippetDataAccess import get_from_dict, fullname
from kamzik3.snippets.snippetsControlLoops import control_device_poller, control_device_connection_poller
from kamzik3.snippets.snippetsGenerators import device_id_generator, token_generator
from kamzik3.snippets.snippetsUnits import device_units
from kamzik3.snippets.snippetsYaml import YamlSerializable


class Device(Subject, YamlSerializable):
    """
    Connection is done in sequence:
        : connect()
        : handle_connect_event()
            : handle_connect()
            : handle_configuration_event()
                : handle_configuration()

    Connection error handling:
        : if connect() => error state
           : disconnect() if critical error
           : cloe(), reconnect() if not critical error
        : if connect() => connection timeout
            : handle_connection_error(), close(), reconnect()
        : if Device.connected and response timeout()
            : handle_response_error(), close_connection(), reconnect()

    Disconnection is done in sequence:
        : disconnect()
            : close() if not Device.connected
            : close_connection() if connected
                : close()
    """
    id_generator = device_id_generator()
    connection_timeout = 4000
    response_timeout = 4000
    allow_reconnect = True
    device_server = None
    session = None
    # Maximum number of commands joined in one request
    max_command_retry = 0
    command_retry_count = 0
    push_commands_max = inf
    push_buffer_size = 2 ** 16

    def __init__(self, device_id=None, config=None):
        Subject.__init__(self)
        self.device_id = device_id
        if device_id is None:
            self.device_id = next(Device.id_generator)
        self.config = config
        if self.config is None:
            self.config = {}
        self.token_generator = token_generator()
        self.device_poller = control_device_poller
        self.device_connection_poller = control_device_connection_poller
        # Get absolute path to the Device class
        self.qualified_name = fullname(self)
        if not hasattr(self, "logger"):
            self.logger = logging.getLogger("Device.{}".format(self.device_id))
        if "push_commands_max" in self.config:
            self.push_commands_max = self.config["push_commands_max"]
        # Size of command buffer used in push
        if "push_buffer_size" in self.config:
            self.push_buffer_size = self.config.get("push_buffer_size")
        self.macro_steps = {MACRO_SET_ATTRIBUTE_STEP: ["*"], MACRO_EXECUTE_METHOD_STEP: ["*"]}
        self.exposed_methods = []
        self._expose_methods_to_clients()
        # Check if attributes are already defined
        if not hasattr(self, "attributes"):
            self.attributes = {}
            self.attributes_sharing_map = {}
            self.attribute_attach_map = {}
            self._init_attributes()
        self.connecting_time = 0
        self.response_timestamp = 0
        self.request_timestamp = 0
        self.connected = False
        self.connecting = False
        self.closing = False
        self.closed = False
        self.connection_error = False
        self.response_error = False
        self.latency_buffer = []
        self.init_timeout = self.init_time = time.time()
        self.commands_buffer = deque()
        if kamzik3.session is not None and self.device_id not in kamzik3.session.devices:
            self.set_session(kamzik3.session)

    def __getitem__(self, item):
        return self.attributes[item]

    def __setitem__(self, key, value):
        self.attributes[key] = value

    def _init_attributes(self):
        """
        Initiate Device's attributes.
        To call this method is mandatory.
        Overload this method for any other derived Device.
        """

        def set_status(key, value):
            if key == VALUE:
                self.notify(ATTR_STATUS, value)

        self.create_attribute(ATTR_ID, default_value=self.device_id, readonly=True, description="Unique device ID")
        self.create_attribute(ATTR_STATUS, default_value=STATUS_DISCONNECTED, readonly=True,
                              description="Current device status").attach_callback(set_status)
        self.create_attribute(ATTR_DESCRIPTION, readonly=True, description="Device description")
        self.create_attribute(ATTR_ENABLED, default_value=True, default_type=np.bool,
                              description="Allow any external changes to device")
        self.create_attribute(ATTR_LATENCY, default_value=0, default_type=np.uint16,
                              description="Latency between sending command and receiving an answer",
                              min_value=0, max_value=9999, unit="ms", readonly=True)
        self.create_attribute(ATTR_BUFFERED_COMMANDS, default_value=0, default_type=np.uint32,
                              description="Amount of commands waiting to be executed",
                              min_value=0, readonly=True)
        self.create_attribute(ATTR_HANGING_COMMANDS, default_value=0, default_type=np.uint32,
                              description="Amount of commands waiting to be answered from device",
                              min_value=0, readonly=True)
        self.create_attribute(ATTR_LAST_ERROR, readonly=True, description="Last device error exception")

    def create_attribute(self, name, group=None, **kwargs):
        attribute_path = name
        if group is not None:
            attribute_path = "{}.{}".format(group, attribute_path)
        attribute_id = "{}.{}.{}".format(self.device_id, TOKEN_ATTRIBUTE, attribute_path)
        attribute = Attribute(attribute_id, **kwargs)

        if group is None:
            self.attributes[name] = attribute
        else:
            if group not in self.attributes:
                self.attributes[group] = {}
            self.attributes[group][name] = attribute
        return attribute

    def add_attribute(self, name, attribute, group=None):
        attribute_list = [name]
        if group is None:
            self.attributes[name] = attribute
        else:
            attribute_list.insert(0, group)
            if group not in self.attributes:
                self.attributes[group] = {}
            self.attributes[group][name] = attribute

        attribute_exists = attribute.attribute_id in self.attributes_sharing_map
        self.attributes_sharing_map[attribute.attribute_id] = attribute_list
        if attribute_exists and self.device_server is not None:
            header = "{}.{}".format(attribute.attribute_id, TOKEN_ATTRIBUTE_REPLACE)
            self.device_server.push_message(header, attribute)
        return attribute

    def share_group(self, source_device, source_group, target_group, attribute_name_mask=None):
        if attribute_name_mask is None:
            attribute_name_mask = {}
        if source_group is not None:
            from_group = source_device.attributes[source_group]
        else:
            from_group = source_device.attributes

        for attribute_name, attribute in from_group.items():
            self.add_attribute(attribute_name_mask.get(attribute_name, attribute_name), attribute, target_group)

    def share_exposed_method(self, source_device, source_method, shared_method_name=None,
                             shared_method_attributes=None):
        """
        Share exposed method with source_device.
        This is handy method when You share attributes of other device and You want to also expose it's methods.
        Each new method will go under new name source_device.id_source_method.name.
        If You define shared_method_attributes they will overwrite default ones.
        :param source_device: Device
        :param source_method: Name of exposed function
        :param shared_method_attributes: Dictionary of method attributes
        :return:
        """
        for method_name, method_parameters in source_device.exposed_methods:
            if method_name == source_method:
                if shared_method_name is None:
                    shared_method_name = "{}_{}".format(source_device.device_id, method_name)
                if shared_method_attributes is None:
                    shared_method_attributes = method_parameters
                method = copy(getattr(source_device, method_name))
                # Check if method is meant to be exposed and remove exposed_parameters to prevent double exposing
                if hasattr(method, "exposed_parameters"):
                    del method.__dict__["exposed_parameters"]
                setattr(self, shared_method_name, method)
                self.exposed_methods.append((shared_method_name, shared_method_attributes))

    def set_session(self, session):
        """
        Set session for device.
        :param session: Session
        :return: None
        """
        self.session = session
        session.add_device(self)

    def reconnect_allowed(self):
        """
        Allow device to reconnect or not.
        :return: Bool
        """
        if self.connecting or self.connected:
            self.close()
            return True
        else:
            return self.allow_reconnect

    def attach_attribute_callback(self, attribute, callback, max_update_rate=None, key_filter=None):
        """
        Attach callback for every change of attribute.
        If max_update_rate is specified that callback will be called only after
        specified timeout.
        :param attribute: [ATTRIBUTE]
        :param callback: callable
        :param max_update_rate: millisecond(s)
        :return: None
        """
        assert callable(callback)
        try:
            attribute_object = self.get_attribute(attribute)
            attribute_object.attach_callback(callback, max_update_rate, key_filter)

            try:
                self.attribute_attach_map[attribute_object.attribute_id].append((attribute, callback))
            except KeyError:
                self.attribute_attach_map.update({attribute_object.attribute_id: [(attribute, callback)]})

            if key_filter is not None:
                callback(attribute_object[key_filter])
            else:
                callback(VALUE, attribute_object[VALUE])
        except (KeyError, AttributeError):
            self.logger.exception(u"Attribute {} is not defined".format(attribute))

    def detach_attribute_callback(self, attribute, callback):
        """
        Detach attached callback from attribute.
        :param attribute: [ATTRIBUTE]
        :param callback: callable
        :return: None
        """
        assert callable(callback)
        try:
            attribute_object = self.get_attribute(attribute)
            attribute_object.detach_callback(callback)

            try:
                if (attribute, callback) in self.attribute_attach_map[attribute_object.attribute_id]:
                    self.attribute_attach_map[attribute_object.attribute_id].remove((attribute, callback))
            except KeyError:
                pass  # Callback does not exists

        except (KeyError, AttributeError):
            if self.logger is not None:
                self.logger.exception(u"Attribute {} is not defined".format(attribute))

    def get_attribute(self, attribute):
        """
        Get Attribute value
        :param attribute: tuple, list, str
        :param value: mixed
        :param callback: callable
        :return: None
        """
        try:
            return self._get(Attribute.list_attribute(attribute))
        except KeyError:
            return None

    def get_value(self, attribute, key=VALUE):
        """
        Get Attribute[key] value
        :param attribute: tuple, list, str
        :param value: mixed
        :param callback: callable
        :return: None
        """
        try:
            return self.get_attribute(attribute)[key]
        except (KeyError, TypeError):
            return None

    def get_raw_value(self, attribute, key=VALUE):
        """
        Get attribute[key] value without offset or factor.
        Use returned value to feed into real hardware.
        :param attribute: list, dict, str
        :return: None
        """
        try:
            attribute_object = self.get_attribute(attribute)
            return attribute_object.remove_offset_factor(attribute_object[key])
        except (KeyError, TypeError):
            return None

    def _get(self, attribute):
        """
        This gets Device attribute.
        Use this function when You want to get attribute by tuple or list key.
        Example: Device.get((ATTR_STATUS, VALUE))
        Otherwise use faster method Device[ATTR_STATUS][VALUE]
        :param attribute: list
        :return: mixed
        """
        try:
            return get_from_dict(self.attributes, attribute)
        except KeyError:
            return None

    def set_attribute(self, attribute, value, callback=None):
        """
        This sets Device attribute.
        Re-implement it for client or any special use.
        :param attribute: tuple, list, str
        :param value: mixed
        :param callback: callable
        :return: None
        """
        list_attribute = Attribute.list_attribute(attribute)
        if list_attribute[-1] == VALUE:
            return self.set_value(list_attribute[:-1], value, callback)
        else:
            return self._set(list_attribute, value, callback)

    def set_value(self, attribute, value, callback=None, key=VALUE):
        """
        Set attribute value.
        :param attribute: list, dict, str
        :param value: mixed
        :return: None
        """
        attribute_list = Attribute.list_attribute(attribute)
        attribute_object = self.get_attribute(attribute_list)
        with SetFunction(attribute_object, callback):
            return self._set(attribute_list + [key], value, callback)

    def set_raw_value(self, attribute, value, callback=None, key=VALUE):
        """
        Set attribute value obtained from device.
        Apply offset and factor before value is set.
        :param attribute: list, dict, str
        :param value: mixed
        :return: None
        """
        list_attribute = Attribute.list_attribute(attribute)
        attribute_object = self.get_attribute(list_attribute)
        value = attribute_object.apply_offset_factor(value)
        self._set(list_attribute + [key], value, callback)

    def _set(self, attribute_list, value, callback=None):
        """
        This sets Device attribute.
        Use this function when You want to set attribute by tuple or list key.
        Example: Device.set((ATTR_STATUS, VALUE), STATUS_IDLE)
        Attribute value is pushed into server if Device is connected listed on one.
        To reduce amount of pushed attributes we check if value is different from previous one.
        :param attribute: list
        :param value: mixed
        :return: None
        """
        attribute_object = get_from_dict(self.attributes, attribute_list[:-1])
        if attribute_list[-1] == VALUE and attribute_object.with_set_function:
            propagate_change = attribute_object.set_value_when_set_function
        else:
            propagate_change = True

        current_value = attribute_object[attribute_list[-1]]
        attribute_object[attribute_list[-1]] = value

        set_token = attribute_object.read_and_reset_token()
        if not set_token and callable(callback):
            callback(attribute_list, value)

        if propagate_change:
            value = attribute_object[attribute_list[-1]]
            if value != current_value and self.device_server is not None:
                header = "{}.{}".format(attribute_object.attribute_id, attribute_list[-1])
                self.device_server.push_message(header, value, min_timeout=attribute_object.min_stream_timeout)

        return set_token

    def connect(self, *args):
        """
        Call only this function to connect devices to port / socket / library / ...
        :param args: connect attributes
        """
        try:
            self.connecting = True
            self.connected = False
            self.device_connection_poller.add_connecting_device(self)
            self.handle_connect_event()
        except DeviceError:
            self.logger.exception(u"Connection exception")
            return

    def handle_readout_callback(self, callback, attribute, output):
        """

        :param callback: callable
        :param attribute: [ATTRIBUTE]
        :param output: mixed
        :return: None
        """
        if callback is None:
            return
        else:
            assert callable(callback)
            callback(attribute, output)

    def handle_connect_event(self):
        """
        Handle connect event.
        This method is a wrapper in which connection is handled.
        Set everything important in handle_connect().
        """
        try:
            self.handle_connect()
            self.connected = True
            self.connecting = False
            self.handle_configuration_event()
        except DeviceError:
            """
            Here we encountered connection error.
            Log error and wait for reconnection.
            """
            self.logger.exception(u"Error during connection")

    def handle_connect(self):
        """
        Load configuration in this method.
        """
        self.set_status(STATUS_CONNECTED)
        self.logger.info(u"Device connection took {} sec.".format(time.time() - self.init_time))
        self.request_timestamp = self.response_timestamp = time.time()

    def handle_configuration_event(self):
        """
        Event from connected device to handle configuration.
        It servers as a wrapper around configuration routine.
        :return:
        """
        try:
            self.set_status(STATUS_CONFIGURING)
            self.handle_configuration()
        except DeviceError:
            self.logger.exception(u"Error during configuration")

    def handle_configuration(self):
        """
        Method is called whenever device is connected and ready to be configured
        :return: None
        """
        raise NotImplementedError(u"Must be implemented in subclass")

    def handle_response_error(self, message=None):
        """
        Method to handle response error of device
        :param message: cause of response error
        :return:
        """
        self.logger.error(message)
        self.response_error = True
        self.close_connection()

    def handle_connection_error(self, message=None):
        """
        Method to handle state of connection error
        :param message: cause of connection error
        :return: None
        """
        self.logger.error(message)
        self.connection_error = True
        self.close()

    def handle_command_error(self, readout_command, readout_output):
        """
        Method to handle state of command error
        :param readout_command: original command
        :param readout_output: error response from device
        :return: None
        """
        self.set_value(ATTR_LAST_ERROR, str(readout_output))
        self.logger.error(
            u"Command error\nCommand: {!r}\nOutput: {!r}\nCommand buffer: {!r}".format(readout_command, readout_output,
                                                                                       self.commands_buffer))

    def handle_observer_attached(self, observer):
        """
        Callback whenever observer is detached from the Subject
        :param observer: Observer
        :return: None
        """
        observer.subject_update(ATTR_STATUS, self.get_value(ATTR_STATUS), self)

    def handle_observer_detached(self, observer):
        """
        Callback whenever observer is attached to the Subject
        :param observer: Observer
        :return:
        """
        pass

    def handle_readout(self, readout_buffer):
        """
        We have data in readoutBuffer.
        Return tuple attribute, joined readoutBuffer
        :param readout_buffer:
        :return: str, str
        """
        self.response_timestamp = time.time()

        try:
            (attribute, token, callback, returning), command_init_timestamp = self.commands_buffer.popleft()
            latency = (time.time() - command_init_timestamp) * 1000
            self.latency_buffer.append(latency)
            if len(self.latency_buffer) == 20:
                self.set_attribute((ATTR_LATENCY, VALUE), sum(self.latency_buffer) / 20.)
                self.set_value(ATTR_HANGING_COMMANDS, len(self.commands_buffer))
                self.latency_buffer = []
        except IndexError:
            self.handle_response_error(
                u"Trying to pop from empty command buffer. Content of readout buffer is: {}".format(readout_buffer))
            return RESPONSE_ERROR, "", None, False

        return attribute, "".join(readout_buffer), callback, token

    def disconnect(self):
        """
        Call this function to cleanly close connection.
        After disconnect Device won't reconnect.
        :return:
        """
        if self.closing:
            return False
        elif self.connected:
            self.close_connection()
        elif self.connecting:
            self.close()

    def close(self):
        """
        Handle all devices closing stuff here.
        Close socket, close port, etc...
        """
        if self.closing:
            return False
        else:
            self.closing = True
            self.connected = False
            self.set_status(STATUS_DISCONNECTED)
            self.closing = False
            self.connected = False

    def reconnect(self):
        """
        Reconnect devices.
        If Device was reconnect return True, False otherwise.
        :return: bool
        """
        if self.reconnect_allowed():
            mapping = self.yaml_mapping()
            self.__init__(**mapping)
            return True
        else:
            return False

    def set_status(self, status):
        """
        Set attributes status value.
        :param status: str
        :return: None
        """
        self.set_value(ATTR_STATUS, status)

    def is_status(self, status):
        """
        Check status value against current device status
        :param status: STATUS
        :return: bool
        """
        return self.get_value(ATTR_STATUS) == status

    def in_statuses(self, statuses):
        """
        Check list of statuses against current device status
        :param statuses: list, tuple
        :return: bool
        """
        return self.get_value(ATTR_STATUS) in statuses

    def set_device_server(self, device_server):
        """
        Set device server
        :param device_server: DeviceServer
        :return: None
        """
        self.device_server = device_server

    def clear_device_server(self):
        self.device_server = None

    def stop_polling(self):
        """
        Remove Device from device_poller loop.
        """
        self.device_poller.stop_polling(self)

    def start_polling(self):
        """
        Add Device attributes to be polled in devicePoller loop.
            self.devicePoller.add(devices, command, time in milliseconds)
            self.devicePoller.add(self, 'cmd0', 400)
        """
        self.response_timestamp = self.request_timestamp = time.time()

    def close_connection(self):
        """
        Call this function when closing connected Devide.
        Stop polling and continue with close() method.
        """
        self.stop_polling()
        self.close()

    def push(self, data):
        """
        Send command to Device.
        This method will send data directly to the Device connection line.
        Use command() for sending commands from clients, etc.
        :param data: str
        """
        pass

    def command(self, command, callback=None, with_token=False, returning=True):
        """
        Send command to devices.
        Use only this function to send command to Device.
        Commands are stored into buffer and are flushed every pollDevice tick.
        :param command:
        :param callback:
        :param with_token:
        :param returning: bool
        :return:
        """
        if not self.valid_command_format(command):
            raise CommandFormatException(u"Command '{}' form is invalid".format(command))

        if self.connected or self.connecting:
            token = 0
            if with_token is True:
                token = next(self.token_generator)
            elif with_token >= 1:
                token = with_token
            self.device_poller.prepare_command(self, (command, token, callback, returning))
            return token

    def remove(self):
        """
        This method aims to remove device from runtime.
        :return:
        """
        self.disconnect()
        if self.device_server is not None:
            self.device_server.remove_device(self)

        if self.logger is not None:
            for handler in self.logger.handlers[:]:
                handler.close()

        self.logger = None
        if self.session is not None:
            self.session.remove_device(self)

        for cb_tuples in self.attribute_attach_map.values():
            for cb_tuple in cb_tuples:
                self.detach_attribute_callback(*cb_tuple)

        self.attribute_attach_map = None
        self.attributes = None
        self.exposed_methods = None
        self.token_generator = None
        self.connecting_time = None
        self.response_timestamp = None
        self.request_timestamp = None
        self.connected = None
        self.connecting = None
        self.closing = None
        self.closed = None
        self.connection_error = None
        self.response_error = None
        self.commands_buffer = None

    def valid_command_format(self, command):
        """
        Check if command format is valid.
        Return False or optionally raise CommandFormatException.
        Re-implement this method for any other device.
        :param command:
        :return:
        """
        if command is not None:
            return True
        else:
            return False

    def send_command(self, commands):
        """
        Send command using defined interface (TCP/IP, RS232, ...).
        Unicode command is always .encode() into ASCII before it's send to the interface.
        :param commands: list of command to be send
        :return: list of NOT sent commands
        """
        if not self.accepting_commands():
            return commands
        try:
            mark = time.time()
            if self.connected:
                command_string = u""
                commands_to_push_counter = 0
                while commands:
                    command = commands.pop(0)
                    commands_to_push_counter += 1
                    command_string += command[0]
                    if len(command_string) > self.push_buffer_size:
                        commands.insert(0, command)
                        command_string = command_string[:-len(command[0])]
                        break
                    if command[3]:  # If command is returning
                        self.commands_buffer.append((command, mark))
                    else:  # Command is not returning, simulate immediate execution
                        self.response_timestamp = mark
                    if commands_to_push_counter >= self.push_commands_max:
                        break
                self.push(command_string.encode())
            else:
                self.handle_response_error(u"Device is not connected")
        except IndexError:
            self.handle_connection_error(u"Device {} buffer error".format(self.device_id))
        except (WriteException, serial.SerialException):
            self.handle_response_error(u"Device {} writing error".format(self.device_id))
        finally:
            return commands

    def accepting_commands(self):
        """
        Check if device is ready to send another commands.
        :return: bool
        """
        self.request_timestamp = time.time()
        if not self.connected or len(self.commands_buffer) > 0:
            return False
        else:
            return True

    def poll_command(self, command, interval):
        """
        Add command to device poller to be polled in desired interval.
        :param command: unicode string
        :param interval: interval in ms
        :return:
        """
        self.device_poller.add(self, command, interval)

    def remove_poll_command(self, command, interval):
        """
        Remove command from device poller.
        :param command:
        :param interval:
        :return:
        """
        self.device_poller.remove(self, command, interval)

    def _config_attributes(self):
        """
        Set attributes to desired values found in config.
        :return: None
        """
        if self.config:
            for attribute, value in self.config.get("attributes", {}).items():
                unit = self.get_attribute(attribute[:-1])[UNIT]
                if unit in (None, "") or attribute[-1] == UNIT:
                    self.set_attribute(attribute, value)
                else:
                    if isinstance(value, list):
                        set_value = []
                        for v in value:
                            try:
                                set_value.append(device_units(self, list(attribute[:-1]), v).m)
                            except (UndefinedUnitError, DimensionalityError):
                                set_value.append(v)
                    else:
                        try:
                            set_value = device_units(self, list(attribute[:-1]), value).m
                        except (UndefinedUnitError, DimensionalityError):
                            set_value = value
                    self.set_attribute(attribute, set_value)

    def _config_commands(self):
        """
        Execute required methods found in commands section of config
        :return: None
        """
        if self.config:
            for command in self.config.get("commands", []):
                self.command(command)

    def _expose_methods_to_clients(self):
        """
        This method is called to expose all marked methods to the clients.
        :return: None
        """
        for method in dir(self):
            method = getattr(self, method)
            if callable(method) and hasattr(method,
                                            "exposed_parameters") and method.__name__ not in self.exposed_methods:
                self.exposed_methods.append((method.__name__, method.exposed_parameters))

    def wait_for_status(self, statuses, retry_timeout=1000, callback=None):
        """
        Wait until device change it's status in one defined in statuses.
        If callback is set, then method is executed in non blocking thread.
        :param statuses: List o one or multiple statuses to check against
        :param retry_timeout: Maximum time in ms to wait for status
        :param callback: Function to call after check in non blocking execution
        :return: True if device is in one of statuses
        """

        def _wait(_callback=None):
            success = True
            retry_counter = 0
            while not self.in_statuses(statuses):
                if not self.connected:
                    return False
                time.sleep(0.05)
                if retry_counter >= retry_timeout:
                    break
                retry_counter += 50
            if _callback is not None:
                _callback(success)
            else:
                if not success:
                    raise DeviceError(
                        "Status was not set to {} within timeout of {} ms".format(statuses, retry_timeout))
                else:
                    return True

        if callback is None:
            return _wait()
        else:
            Thread(target=_wait, args=[callback]).start()

    def wait_until_value_set(self, attribute, value, retry_timeout=1000, callback=None):

        def _wait(_callback=None):
            attribute_object = self.get_attribute(attribute)
            success = True
            retry_counter = 0
            while attribute_object[VALUE] != value:
                if not self.connected:
                    return False
                time.sleep(0.05)
                if retry_counter >= retry_timeout:
                    success = False
                    break
                retry_counter += 50
            if _callback is not None:
                _callback(success)
            else:
                if not success:
                    raise DeviceError(
                        "Value of {} was not set to {} within timeout of {} ms".format(attribute, value, retry_timeout))
                else:
                    return True

        if callback is None:
            return _wait()
        else:
            thread = Thread(target=_wait, args=[callback])
            thread.start()
            return thread

    def is_alive(self):
        if not self.connected or self.closing or self.response_error:
            return False
        time_diff = self.request_timestamp - self.response_timestamp
        if time_diff >= self.response_timeout * 1e-3:
            print(self.device_id, time_diff, self.request_timestamp, self.response_timestamp, self.response_timeout * 1e-3)
            print(len(self.commands_buffer), self.commands_buffer)
            self.handle_response_error(u"Response timeout")
            return False
        return True
